# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from __future__ import annotations

import copy
import csv
import json
import os
import shutil
from os.path import isfile, join
from typing import Any, Dict, Optional, Union

import pcre as re
import torch
import transformers
from safetensors import safe_open
from safetensors.torch import save_file
from transformers import AutoConfig, PreTrainedTokenizerFast, ProcessorMixin
from transformers.models.auto.tokenization_auto import get_tokenizer_config
from transformers.utils.generic import ContextManagers

from ..adapter.adapter import HF_ADAPTER_FILE_NAME, HF_ADAPTER_WEIGHT_KEY_PREFIX, Lora
from ..adapter.peft import LoraConfig
from ..quantization.config import (
    FORMAT,
    META_FIELD_ACT_GROUP_AWARE,
    META_FIELD_DAMP_AUTO_INCREMENT,
    META_FIELD_DAMP_PERCENT,
    META_FIELD_GPTAQ_ENABLED,
    META_FIELD_MSE,
    META_FIELD_QUANTIZER,
    META_FIELD_STATIC_GROUPS,
    META_FIELD_TRUE_SEQUENTIAL,
    META_FIELD_URI,
    META_QUANTIZER_GPTQMODEL,
    META_VALUE_URI,
    MIN_VERSION_WITH_V2,
)
from ..utils.backend import BACKEND
from ..utils.hf import no_init_weights, prepare_remote_code_compat, sanitize_generation_config_file, \
    sanitize_model_config
from ..utils.logger import setup_logger
from ..utils.model import (
    copy_py_files,
    find_modules,
    get_model_files_size,
    get_state_dict_for_save,
    load_checkpoint_in_model_then_tie_weights,
    make_quant,
    streaming_state_dict_to_shards,
)
from ..utils.structure import alias_all_from_turtle_if_meta
from ..utils.torch import torch_empty_cache
from ..version import __version__
from ._const import DEFAULT_MAX_SHARD_SIZE, DEVICE


log = setup_logger()

PROCESS_LOG_NAME = "process"
PROCESS_LOG_LAYER = "layer"
PROCESS_LOG_MODULE = "module"
QUANT_LOG_LOSS = "loss"
QUANT_LOG_NSAMPLES = "samples"
QUANT_LOG_DAMP = "damp"
PROCESS_LOG_TIME = "time"
PROCESS_LOG_FWD_TIME = "fwd_time"
PROCESS_USED_MEMORY = "(v)ram"

EORA_DEFAULT_FILE = "eora.safetensors"

def ModelWriter(cls):
    def save_pretrained(
            self,
            save_dir: str,
            **kwargs,
    ):
        log.warn("You are using save_pretrained, which will re-direct to save_quantized.")
        self.save_quantized(save_dir=save_dir, **kwargs)

    cls.save_pretrained = save_pretrained

    def _eora_save(self, save_dir: str, model_save_dir: str = None):
        assert isinstance(self.quantize_config.adapter, Lora)

        assert hasattr(self, 'lora_results')

        # save lora tensors
        if self.lora_results:  # TODO REFRACTOR
            weights = {}
            target_modules = set()
            # convert the dict into safetensors compatible dict
            for key, adapter in self.lora_results.items():
                assert isinstance(adapter, Lora)
                key = key.lower()
                simple_module_name = key.split(".")[-1] # mlp.gate_proj => gate_proj
                target_modules.add(simple_module_name)

                # while key.startswith('model.'):
                #     key = key.removeprefix('model.') # some HF models use model. or model.model.

                # must normalize key since HF can load weights as `model.` or not based on what AutoModel is used
                weight_key = f"{HF_ADAPTER_WEIGHT_KEY_PREFIX}{key}"

                weights[f"{weight_key}.lora_A.weight"] = adapter.lora_A
                weights[f"{weight_key}.lora_B.weight"] = adapter.lora_B
                log.info(f"Adapter: EoRA weights found -> `{weight_key}.lora_A/Lora_B.weight`, rank = `{adapter.rank}`")

            weight_file_path = f"{save_dir.removesuffix('/')}/{HF_ADAPTER_FILE_NAME}"

            # dynamic rank
            rank_pattern = {}
            if self.quantize_config.dynamic:
                rank_pattern = self.quantize_config.extract_adapter_rank_patterns()

            lora_cfg = LoraConfig(base_model_name_or_path=model_save_dir,
                                  r=self.quantize_config.adapter.rank,
                                  lora_alpha=self.quantize_config.adapter.rank,
                                  target_modules=list(target_modules),
                                  rank_pattern=rank_pattern)
            lora_cfg.save_pretrained(save_dir=save_dir)

            log.info(f"Adapter: Saving EoRA weights to -> `{save_dir}`")

            save_file(tensors=weights, filename=weight_file_path, metadata={"format": "pt"})

            del self.lora_results  # TODO REFRACTOR

    cls.eora_save = _eora_save

    def save_quantized(
            self,
            save_dir: str,
            safetensors_metadata: Optional[Dict[str, str]] = None,
            max_shard_size: Optional[Union[int, str]] = DEFAULT_MAX_SHARD_SIZE,
            meta_quantizer: Optional[str] = None,
            eora_path: Optional[str] = None,
    ):
        """save quantized model and configs to local disk"""
        os.makedirs(save_dir, exist_ok=True)

        if self.quant_log:
            with open(os.path.join(save_dir, "quant_log.csv"), mode='w', newline='') as file:
                w = csv.writer(file)
                w.writerow([PROCESS_LOG_LAYER, PROCESS_LOG_MODULE, QUANT_LOG_LOSS, QUANT_LOG_NSAMPLES, QUANT_LOG_DAMP, PROCESS_LOG_TIME])
                w.writerows([[entry.get(PROCESS_LOG_LAYER), entry.get(PROCESS_LOG_MODULE), entry.get(QUANT_LOG_LOSS),
                              entry.get(QUANT_LOG_DAMP), entry.get(PROCESS_LOG_TIME)] for entry in self.quant_log])

        pre_quantized_size_mb = get_model_files_size(self.model_local_path)
        pre_quantized_size_gb = pre_quantized_size_mb / 1024

        quantizers = [f"{META_QUANTIZER_GPTQMODEL}:{__version__}"]
        if meta_quantizer:
            if len(meta_quantizer.split(":")) == 2:
                quantizers.append(meta_quantizer.replace(" ",""))
            else:
                log.warn(f"meta_quantizer: '{meta_quantizer}' format is invalid, expected: 'quantizer_name:version'")

        # write gptqmodel tooling fingerprint to config
        self.quantize_config.meta_set_versionable(
            key=META_FIELD_QUANTIZER,
            value=quantizers
        )


        self.quantize_config.meta_set(
            key=META_FIELD_URI,
            value=META_VALUE_URI,
        )

        # meta: write config fields to meta if they doe not participate in inference
        self.quantize_config.meta_set(
            key=META_FIELD_DAMP_PERCENT,
            value=self.quantize_config.damp_percent
        )

        self.quantize_config.meta_set(
            key=META_FIELD_DAMP_AUTO_INCREMENT,
            value=self.quantize_config.damp_auto_increment
        )

        self.quantize_config.meta_set(
            key=META_FIELD_STATIC_GROUPS,
            value=self.quantize_config.static_groups
        )

        self.quantize_config.meta_set(
            key=META_FIELD_TRUE_SEQUENTIAL,
            value=self.quantize_config.true_sequential
        )

        self.quantize_config.meta_set(
            key=META_FIELD_MSE,
            value=self.quantize_config.mse
        )

        self.quantize_config.meta_set(
            key=META_FIELD_GPTAQ_ENABLED,
            value=None if self.quantize_config.gptaq is None else {
                "alpha": self.quantize_config.gptaq.alpha,
                "device": (
                    self.quantize_config.gptaq.device
                    if isinstance(self.quantize_config.gptaq.device, str)
                    else str(self.quantize_config.gptaq.device)
                ),
            }
        )

        self.quantize_config.meta_set(
            key=META_FIELD_ACT_GROUP_AWARE,
            value=self.quantize_config.act_group_aware
        )

        # The config, quantize_config and model may be edited in place in save_quantized.
        sanitize_model_config(self.model.config)
        config = copy.deepcopy(self.model.config)

        quantize_config = copy.deepcopy(self.quantize_config)

        if not self.quantized:
            raise ValueError("Save aborted as model is not quantized. Please call `quantize()` first.")

        if quantize_config.format == FORMAT.GPTQ_V2:
            log.warn(
                f"Using 'format = {FORMAT.GPTQ_V2}': the serialized model is only supported by GPTQModel version >= {MIN_VERSION_WITH_V2}."
            )

        if self.load_quantized_model:
            self.model = self.get_model_with_quantize(
                qcfg=quantize_config,
                model_id_or_path=self.model_local_path,
            )

        # --- start config save block ---
        # Save quantized config
        config.quantization_config = quantize_config.to_dict()
        self.model.config = config

        def strip_attention_impl_fields(target: Any) -> Dict[str, Any]:
            removed: Dict[str, Any] = {}
            for attr in ("attn_implementation", "_attn_implementation"):
                if hasattr(target, attr):
                    removed[attr] = getattr(target, attr)
                    # Avoid AttributeError: property '_attn_implementation' of 'Qwen2Config' object has no deleter
                    try:
                        delattr(target, attr)
                    except Exception:
                        pass
            return removed

        generation_config = getattr(self.model, "generation_config", None)
        removed_config_attention_attrs: Dict[str, Any] = {}
        removed_generation_attention_attrs: Dict[str, Any] = {}

        try:
            removed_config_attention_attrs = strip_attention_impl_fields(self.model.config)
            if generation_config is not None:
                removed_generation_attention_attrs = strip_attention_impl_fields(generation_config)

            # Save model config, including generation_config
            # Use empty state_dict hack to bypass saving weights
            self.model.save_pretrained(save_dir, state_dict={}, is_main_process=True)
        finally:
            for attr, value in removed_config_attention_attrs.items():
                setattr(self.model.config, attr, value)
            if generation_config is not None:
                for attr, value in removed_generation_attention_attrs.items():
                    setattr(generation_config, attr, value)

        gen_config_path = os.path.join(save_dir, "generation_config.json")
        if sanitize_generation_config_file(gen_config_path):
            log.info("Model: Sanitized `generation_config.json` before packaging.")

        # Save `quantize_config.json`
        quantize_config.save_pretrained(save_dir)

        def debug_saved_config(path):
            # List all files in the directory
            files = os.listdir(path)
            print("Files in directory:")
            for file in files:
                print(file)

            config_file_paths = ["generation_config.json", "config.json"]
            for file_name in config_file_paths:
                full_path = os.path.join(path, file_name)
                if os.path.isfile(full_path):
                    print(f"Content of saved `{file_name}`:")
                    with open(full_path, 'r') as config_file:
                        config_data = json.load(config_file)
                        print(json.dumps(config_data, indent=4))
                else:
                    print(f"`{file_name}` does not exist in the directory.")

        debug_saved_config(save_dir)

        # Save processor related config files. For example: preprocessor_config.json, chat_template.json
        if hasattr(self,"processor") and isinstance(self.processor, ProcessorMixin):
            self.processor.save_pretrained(save_dir)
        # --- end config save block ---

        # Due to shell/turtle state, we need to sync the modules from turtle to shell
        if not self.load_quantized_model:
            alias_all_from_turtle_if_meta(shell_model=self.model, turtle_model=self.turtle_model)

        offload_root = self.quantize_config.offload_to_disk_path if getattr(self.quantize_config, "offload_to_disk", False) else None
        state_dict = get_state_dict_for_save(self.model, offload_root=offload_root)

        model_base_name = "model"
        model_save_name = model_base_name + ".safetensors"

        if not self.qlinear_kernel.SUPPORTS_SHARDS and max_shard_size is not None:
            log.warn("Sharding is not supported for this quant. Disabling sharding.")
            max_shard_size = None

        def _parse_max_shard_size(value: Optional[Union[int, str]]) -> Optional[int]:
            if value is None:
                return None
            if isinstance(value, int):
                return value
            match = re.fullmatch(r"\s*(\d+)([KMGTP]?B?)\s*", value, re.IGNORECASE)
            if not match:
                raise ValueError(f"Invalid max_shard_size value: {value}")
            base = int(match.group(1))
            suffix = match.group(2).upper()
            multiplier = 1
            if suffix.startswith("K"):
                multiplier = 1024
            elif suffix.startswith("M"):
                multiplier = 1024 ** 2
            elif suffix.startswith("G"):
                multiplier = 1024 ** 3
            elif suffix.startswith("T"):
                multiplier = 1024 ** 4
            elif suffix.startswith("P"):
                multiplier = 1024 ** 5
            return base * multiplier

        def _normalize_metadata(meta: Optional[Dict[str, Any]]) -> Dict[str, str]:
            if meta is None:
                return {}
            if not isinstance(meta, dict):
                raise TypeError("safetensors_metadata must be a dictionary.")
            normalized: Dict[str, str] = {}
            for key, value in meta.items():
                try:
                    new_key = str(key)
                    new_value = str(value)
                except Exception as exc:
                    raise TypeError(
                        f"safetensors_metadata: both keys and values must be strings and conversion failed for ({key}, {value}): {exc}"
                    )
                if new_key in normalized:
                    log.warn(
                        f"Duplicate metadata key '{new_key}' after conversion to string; overwriting previous value."
                    )
                normalized[new_key] = new_value
            return normalized

        max_shard_size_bytes = _parse_max_shard_size(max_shard_size)
        metadata_dict = _normalize_metadata(safetensors_metadata)
        metadata_dict["format"] = "pt"

        expected_files, tensor_to_filename, total_size_bytes = streaming_state_dict_to_shards(
            state_dict,
            save_dir=save_dir,
            model_base_name=model_base_name,
            single_file_name=model_save_name,
            metadata=metadata_dict,
            max_shard_size=max_shard_size_bytes,
        )

        pattern = re.compile(rf"{re.escape(model_base_name)}-\d{{5}}-of-\d{{5}}\.safetensors")
        for filename in os.listdir(save_dir):
            full_filename = join(save_dir, filename)
            if not isfile(full_filename):
                continue
            if filename == model_save_name and filename not in expected_files:
                os.remove(full_filename)
                continue
            if pattern.fullmatch(filename) and filename not in expected_files:
                os.remove(full_filename)

        total_size_mb = total_size_bytes / (1024 * 1024)

        if len(expected_files) > 1:
            index = {
                "metadata": {"total_size": total_size_bytes},
                "weight_map": tensor_to_filename,
            }
            index_save_name = model_save_name + ".index.json"
            index_save_path = join(save_dir, index_save_name)
            with open(index_save_path, "w", encoding="utf-8") as f:
                content = json.dumps(index, indent=2, sort_keys=True) + "\n"
                f.write(content)
        else:
            index_save_path = join(save_dir, model_save_name + ".index.json")
            if os.path.exists(index_save_path):
                os.remove(index_save_path)

        state_dict.clear()

        # save lora
        if self.quantize_config.adapter:
            _eora_save(self, save_dir=eora_path if eora_path else self.quantize_config.adapter.path, model_save_dir=save_dir)

        # Handle `dangling` tensor files that HF doesn't support (optional) but very useful
        extra_tensor_files = getattr(self, "out_of_model_tensor_files", None)
        if extra_tensor_files:
            if isinstance(extra_tensor_files, str):
                extra_tensor_files = [extra_tensor_files]
            else:
                extra_tensor_files = list(extra_tensor_files)

            index_save_name = model_save_name + ".index.json"
            index_save_path = join(save_dir, index_save_name)

            if os.path.exists(index_save_path):
                with open(index_save_path, "r", encoding="utf-8") as f:
                    index_data = json.load(f)
            else:
                index_data = {
                    "metadata": {"total_size": total_size_bytes},
                    "weight_map": dict(tensor_to_filename),
                }

            if "metadata" not in index_data:
                index_data["metadata"] = {}
            if "weight_map" not in index_data:
                index_data["weight_map"] = {}

            total_size_value = index_data["metadata"].get("total_size", total_size_bytes)
            index_updated = False

            for tensor_file_name in extra_tensor_files:
                original_tensor_path = os.path.join(self.model_local_path, tensor_file_name)
                if not os.path.exists(original_tensor_path):
                    log.warn(
                        f"Model: out_of_model_tensor_files configured with '{tensor_file_name}', "
                        f"but the file was not found at '{original_tensor_path}'"
                    )
                    continue

                target_tensor_path = os.path.join(save_dir, tensor_file_name)
                shutil.copy2(original_tensor_path, target_tensor_path)
                log.info(
                    f"Model: Copied {tensor_file_name} from original model directory to quantized model directory"
                )

                tensor_names = []
                try:
                    with safe_open(original_tensor_path, framework="pt", device="cpu") as f:
                        tensor_names = list(f.keys())
                except Exception as exc:
                    log.warn(
                        f"Model: Failed to read tensor names from {tensor_file_name}: {exc}"
                    )

                for tensor_name in tensor_names:
                    index_data["weight_map"][tensor_name] = tensor_file_name

                if tensor_names:
                    log.info(
                        f"Model: Added {len(tensor_names)} tensors from {tensor_file_name} to weight_map"
                    )

                try:
                    tensor_file_size = os.path.getsize(target_tensor_path)
                except OSError:
                    tensor_file_size = 0

                total_size_value += tensor_file_size
                index_updated = True

            if index_updated:
                index_data["metadata"]["total_size"] = total_size_value
                with open(index_save_path, "w", encoding="utf-8") as f:
                    content = json.dumps(index_data, indent=2, sort_keys=True) + "\n"
                    f.write(content)
                log.info(
                    f"Model: Updated {index_save_name} to include `out_of_model_tensor_files`"
                )

        # If the saved model is a loaded quantized model, do not calculate the size diff.
        if not self.load_quantized_model:
            total_size_gb = total_size_mb / 1024
            size_diff_mb = pre_quantized_size_mb - total_size_mb
            size_diff_gb = size_diff_mb / 1024
            percent_diff = (size_diff_mb / pre_quantized_size_mb) * 100
            log.info(f"Pre-Quantized model size: {pre_quantized_size_mb:.2f}MB, {pre_quantized_size_gb:.2f}GB")
            log.info(f"Quantized model size: {total_size_mb:.2f}MB, {total_size_gb:.2f}GB")
            log.info(f"Size difference: {size_diff_mb:.2f}MB, {size_diff_gb:.2f}GB - {percent_diff:.2f}%")

        # need to copy .py files for model/tokenizers not yet merged to HF transformers
        if self.trust_remote_code:
            copy_py_files(save_dir, model_id_or_path=self.model_local_path)

        if self.tokenizer:
            self.tokenizer.save_pretrained(save_dir)

            # Use source model's tokenizer_class for cross-version compatibility
            # (transformers 5.0 save_pretrained writes "TokenizersBackend" which older versions can't load)
            source_tokenizer_config = get_tokenizer_config(self.model_local_path)
            source_tokenizer_class = source_tokenizer_config.get("tokenizer_class")

            if source_tokenizer_class:
                # fix https://github.com/huggingface/transformers/issues/35832
                # if source tokenizer_class lacks "Fast" suffix but tokenizer is actually fast, add it
                if (not source_tokenizer_class.endswith("Fast")) and isinstance(self.tokenizer.tokenizer, PreTrainedTokenizerFast):
                    source_tokenizer_class = source_tokenizer_class + "Fast"

                saved_tokenizer_config = get_tokenizer_config(save_dir)
                if saved_tokenizer_config.get("tokenizer_class") != source_tokenizer_class:
                    saved_tokenizer_config["tokenizer_class"] = source_tokenizer_class
                    with open(os.path.join(save_dir, "tokenizer_config.json"), "w", encoding="utf-8") as f:
                        json.dump(saved_tokenizer_config, f, indent=2, ensure_ascii=False)


    cls.save_quantized = save_quantized

    def get_model_with_quantize(self, qcfg, model_id_or_path):

        config = AutoConfig.from_pretrained(
            model_id_or_path,
            trust_remote_code=True,
        )
        prepare_remote_code_compat(config)

        def skip(*args, **kwargs):
            pass

        torch.nn.init.kaiming_uniform_ = skip
        torch.nn.init.uniform_ = skip
        torch.nn.init.normal_ = skip
        transformers.modeling_utils._init_weights = False
        init_contexts = [no_init_weights()]
        with ContextManagers(init_contexts):
            model = cls.loader.from_config(
                config, dtype=torch.float16
            )

            modules = find_modules(model)
            ignore_modules = [self.lm_head] + self.get_base_modules(model)

            for name in list(modules.keys()):
                # allow loading of quantized lm_head
                if qcfg.lm_head and name == self.lm_head:
                    continue

                if any(name.startswith(ignore_module) for ignore_module in ignore_modules) or all(
                        not name.endswith(ignore_module) for sublist in self.simple_layer_modules(config, qcfg) for ignore_module in sublist
                ):
                    # log non-lm-head quantizerd modules only
                    if name is not self.lm_head:
                        log.info(f"The layer {name} is not quantized.")
                    del modules[name]


            make_quant(
                model,
                qcfg=qcfg,
                quant_result=modules,
                backend=BACKEND.AUTO,
                lm_head_name=cls.lm_head,
                pack=True,
                device=DEVICE.CPU,
            )

        load_checkpoint_in_model_then_tie_weights(
            model,
            dtype=torch.float16,
            # This is very hacky but works due to https://github.com/huggingface/accelerate/blob/bd72a5f1a80d5146554458823f8aeda0a9db5297/src/accelerate/utils/modeling.py#L292
            checkpoint=self.checkpoint_file_name,
            # device_map=device_map,
            # offload_state_dict=True,
            # offload_buffers=True,
        )
        torch_empty_cache()
        return model

    cls.get_model_with_quantize = get_model_with_quantize

    return cls



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from __future__ import annotations

import os

from ..utils.logger import setup_logger


log = setup_logger()

ASCII_LOGO = r"""
┌─────────────┐    ┌────────────────────────┐    ┌────────────┐    ┌─────────┐
│ GPT-QModel  │ -> │ ▓▓▓▓▓▓▓▓▓▓▓▓ 16bit     │ -> │ ▒▒▒▒ 8bit  │ -> │ ░░ 4bit │
└─────────────┘    └────────────────────────┘    └────────────┘    └─────────┘
"""

# if not os.environ.get("PYTHON_GIL", None):
#     os.environ["PYTHON_GIL"] = '0'
#     log.info("ENV: Auto disable GIL and use free-threading mode when applicable: Python 3.13t+. You must install the -t edition of Python.")

if not os.environ.get("PYTORCH_ALLOC_CONF", None):
    os.environ["PYTORCH_ALLOC_CONF"] = 'expandable_segments:True,max_split_size_mb:256,garbage_collection_threshold:0.7'
    log.info("ENV: Auto setting PYTORCH_ALLOC_CONF='expandable_segments:True,max_split_size_mb:256,garbage_collection_threshold:0.7' for memory saving.")

if not os.environ.get("CUDA_DEVICE_ORDER", None):
    os.environ["CUDA_DEVICE_ORDER"] = 'PCI_BUS_ID'
    log.info("ENV: Auto setting CUDA_DEVICE_ORDER=PCI_BUS_ID for correctness.")

# FIX ROCm env conflict with CUDA_VISIBLE_DEVICES if both exits
if 'CUDA_VISIBLE_DEVICES' in os.environ and 'ROCR_VISIBLE_DEVICES' in os.environ:
    del os.environ['ROCR_VISIBLE_DEVICES']

# if not os.environ.get("NCCL_SHM_DISABLE", None):
#     os.environ["NCCL_SHM_DISABLE"] = '1'
#     log.info("ENV: Auto setting NCCL_SHM_DISABLE=1 for multi-gpu memory safety.")

import sys  # noqa: E402


# TODO: waiting for pytorch implementgation of aten ops for MPS
if sys.platform == "darwin":
    os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

import os.path  # noqa: E402
import random  # noqa: E402
from os.path import isdir, join  # noqa: E402
from typing import Any, Dict, List, Optional, Type, Union  # noqa: E402

import numpy  # noqa: E402
import torch  # noqa: E402
from huggingface_hub import list_repo_files  # noqa: E402
from packaging.version import Version  # noqa: E402
from tokenicer import Tokenicer  # noqa: E402
from transformers import AutoConfig, GenerationConfig, PreTrainedModel, PreTrainedTokenizerBase  # noqa: E402
from transformers import __version__ as TRANSFORMERS_VERSION

from ..adapter.adapter import Adapter, Lora, normalize_adapter  # noqa: E402
from ..nn_modules.qlinear.torch import TorchQuantLinear  # noqa: E402
from ..quantization import METHOD, QUANT_CONFIG_FILENAME  # noqa: E402
from ..utils import BACKEND  # noqa: E402
from ..utils.eval import EVAL  # noqa: E402
from ..utils.hf import resolve_trust_remote_code  # noqa: E402
from ..utils.model import find_modules  # noqa: E402
from ..utils.torch import CPU, torch_empty_cache  # noqa: E402
from .base import BaseQModel, QuantizeConfig  # noqa: E402
from .definitions.afmoe import AfMoeQModel  # noqa: E402
from .definitions.apertus import ApertusQModel  # noqa: E402
from .definitions.baichuan import BaiChuanQModel  # noqa: E402
from .definitions.bailing_moe import BailingMoeQModel  # noqa: E402
from .definitions.bloom import BloomQModel  # noqa: E402
from .definitions.brumby import BrumbyQModel  # noqa: E402
from .definitions.chatglm import ChatGLMQModel  # noqa: E402
from .definitions.codegen import CodeGenQModel  # noqa: E402
from .definitions.dbrx import DbrxQModel  # noqa: E402
from .definitions.dbrx_converted import DbrxConvertedQModel  # noqa: E402
from .definitions.decilm import DeciLMQModel  # noqa: E402
from .definitions.deepseek_v2 import DeepSeekV2QModel  # noqa: E402
from .definitions.deepseek_v3 import DeepSeekV3QModel  # noqa: E402
from .definitions.dots1 import Dots1QModel  # noqa: E402
from .definitions.dream import DreamQModel  # noqa: E402
from .definitions.ernie4_5 import Ernie4_5QModel  # noqa: E402
from .definitions.ernie4_5_moe import Ernie4_5_MoeQModel  # noqa: E402
from .definitions.exaone import ExaOneQModel  # noqa: E402
from .definitions.exaone4 import Exaone4QModel  # noqa: E402
from .definitions.falcon_h1 import FalconH1QModel  # noqa: E402
from .definitions.gemma2 import Gemma2QModel  # noqa: E402
from .definitions.gemma3 import Gemma3ForConditionalGenerationGPTQ, Gemma3QModel  # noqa: E402
from .definitions.glm import GlmQModel  # noqa: E402
from .definitions.glm4_moe import GLM4MoEGPTQ  # noqa: E402
from .definitions.glm4v import Glm4vGPTQ  # noqa: E402
from .definitions.gpt2 import GPT2QModel  # noqa: E402
from .definitions.gpt_bigcode import GptBigCodeQModel  # noqa: E402
from .definitions.gpt_neo import GptNeoQModel  # noqa: E402
from .definitions.gpt_neox import GPTNeoXQModel  # noqa: E402
from .definitions.gpt_oss import GPTOSSGPTQ  # noqa: E402
from .definitions.gptj import GptJQModel  # noqa: E402
from .definitions.granitemoehybrid import GraniteMoeHybridQModel
from .definitions.grinmoe import GrinMoeQModel  # noqa: E402
from .definitions.hymba import HymbaQModel  # noqa: E402
from .definitions.instella import InstellaQModel  # noqa: E402
from .definitions.internlm import InternLMQModel  # noqa: E402
from .definitions.internlm2 import InternLM2QModel  # noqa: E402
from .definitions.klear import KlearQModel  # noqa: E402
from .definitions.lfm2_moe import LFM2MoeQModel  # noqa: E402
from .definitions.llada2 import LLaDA2MoeQModel
from .definitions.llama import LlamaQModel  # noqa: E402
from .definitions.llama4 import Llama4QModel  # noqa: E402
from .definitions.llava_qwen2 import LlavaQwen2QModel  # noqa: E402
from .definitions.longcat_flash import LongCatFlashQModel  # noqa: E402
from .definitions.mimo import MimoQModel  # noqa: E402
from .definitions.minicpm import MiniCPMGPTQ  # noqa: E402
from .definitions.minicpm3 import MiniCpm3QModel  # noqa: E402
from .definitions.minimax_m2 import MiniMaxM2GPTQ  # noqa: E402
from .definitions.mistral3 import Mistral3GPTQ
from .definitions.mixtral import MixtralQModel  # noqa: E402
from .definitions.mllama import MLlamaQModel  # noqa: E402
from .definitions.mobilellm import MobileLLMQModel  # noqa: E402
from .definitions.moss import MossQModel  # noqa: E402
from .definitions.mpt import MptQModel  # noqa: E402
from .definitions.nemotron_h import NemotronHQModel  # noqa: E402
from .definitions.opt import OptQModel  # noqa: E402
from .definitions.ovis import OvisQModel  # noqa: E402
from .definitions.ovis2 import Ovis2QModel  # noqa: E402
from .definitions.pangu_alpha import PanguAlphaQModel  # noqa: E402
from .definitions.phi import PhiQModel  # noqa: E402
from .definitions.phi3 import Phi3QModel, PhiMoEGPTQForCausalLM  # noqa: E402
from .definitions.phi4 import Phi4MMGPTQ  # noqa: E402
from .definitions.qwen import QwenQModel  # noqa: E402
from .definitions.qwen2 import Qwen2QModel  # noqa: E402
from .definitions.qwen2_5_omni import Qwen2_5_OmniGPTQ
from .definitions.qwen2_5_vl import Qwen2_5_VLQModel  # noqa: E402
from .definitions.qwen2_moe import Qwen2MoeQModel  # noqa: E402
from .definitions.qwen2_vl import Qwen2VLQModel  # noqa: E402
from .definitions.qwen3 import Qwen3QModel  # noqa: E402
from .definitions.qwen3_moe import Qwen3MoeQModel  # noqa: E402
from .definitions.qwen3_next import Qwen3NextGPTQ  # noqa: E402
from .definitions.qwen3_omni_moe import Qwen3OmniMoeGPTQ
from .definitions.qwen3_vl import Qwen3_VLQModel
from .definitions.rw import RwgQModel  # noqa: E402
from .definitions.starcoder2 import Starcoder2QModel  # noqa: E402
from .definitions.telechat2 import TeleChat2QModel
from .definitions.voxtral import VoxtralGPTQ  # noqa: E402
from .definitions.xverse import XverseQModel  # noqa: E402


TRANSFORMERS_SUPPORTS_QWEN3_5 = Version(TRANSFORMERS_VERSION) >= Version("5.2.0")
if TRANSFORMERS_SUPPORTS_QWEN3_5:
    from .definitions.qwen3_5 import Qwen3_5QModel  # noqa: E402
    from .definitions.qwen3_5_moe import Qwen3_5_MoeQModel  # noqa: E402
else:
    Qwen3_5QModel = None
    Qwen3_5_MoeQModel = None


# make quants and inference more determinisitc
torch.manual_seed(787)
random.seed(787)
numpy.random.seed(787)

MODEL_MAP = {
    "apertus": ApertusQModel,
    "dream": DreamQModel,
    "bloom": BloomQModel,
    "brumby": BrumbyQModel,
    "gpt_neo": GptNeoQModel,
    "kimi_k2": DeepSeekV3QModel, # 100% DeepSeekV3QModel clone
    "klear": KlearQModel,
    "gpt_neox": GPTNeoXQModel,
    "gptj": GptJQModel,
    "gpt2": GPT2QModel,
    "llama": LlamaQModel,
    "llama4": Llama4QModel,
    "opt": OptQModel,
    "moss": MossQModel,
    "chatglm": ChatGLMQModel,
    "glm": GlmQModel,
    "glm4": GlmQModel,
    "glm4v": Glm4vGPTQ,
    "glm4_moe": GLM4MoEGPTQ,
    "gpt_bigcode": GptBigCodeQModel,
    "codegen": CodeGenQModel,
    "cohere": LlamaQModel, # 100% llama clone
    "cohere2": LlamaQModel, # 100% llama clone
    "refinedWebModel": RwgQModel,
    "refinedWeb": RwgQModel,
    "falcon": RwgQModel,
    "baichuan": BaiChuanQModel,
    "internlm": InternLMQModel,
    "internlm2": InternLM2QModel,
    "qwen": QwenQModel,
    "mistral": LlamaQModel, # 100% llama clone
    "yi": LlamaQModel, # 100% llama clone
    "xverse": XverseQModel,
    "deci": DeciLMQModel,
    "nemotron-nas": DeciLMQModel,
    "stablelm_epoch": LlamaQModel, # 100% llama clone
    "stablelm": LlamaQModel, # 100% llama clone
    "starcoder2": Starcoder2QModel,
    "mixtral": MixtralQModel,
    "qwen2": Qwen2QModel,
    "qwen3": Qwen3QModel,
    "longllama": LlamaQModel,  # 100% llama clone
    "gemma": LlamaQModel, # 100% llama clone
    "gemma2": Gemma2QModel,
    "gemma3_text": Gemma3QModel,
    "gemma3": Gemma3ForConditionalGenerationGPTQ,
    "phi": PhiQModel,
    "phi3": Phi3QModel,
    "phi4mm": Phi4MMGPTQ,
    "phimoe": PhiMoEGPTQForCausalLM,
    "mpt": MptQModel,
    "minicpm": MiniCPMGPTQ,
    "minicpm3": MiniCpm3QModel,
    "minimax": MiniMaxM2GPTQ,
    "minimax_m2": MiniMaxM2GPTQ,
    "qwen2_moe": Qwen2MoeQModel,
    "qwen3_moe": Qwen3MoeQModel,
    "qwen3_next": Qwen3NextGPTQ,
    "qwen2_vl": Qwen2VLQModel,
    "qwen2_vl_text": Qwen2VLQModel,
    "qwen2_5_vl": Qwen2_5_VLQModel,
    "qwen2_5_vl_text": Qwen2_5_VLQModel,
    "qwen2_5_omni": Qwen2_5_OmniGPTQ,
    "qwen3_omni_moe": Qwen3OmniMoeGPTQ,
    "qwen3_vl": Qwen3_VLQModel,
    "dbrx": DbrxQModel,
    "dbrx_converted": DbrxConvertedQModel,
    "deepseek_v2": DeepSeekV2QModel,
    "deepseek_v3": DeepSeekV3QModel,
    "dots1": Dots1QModel,
    "exaone": ExaOneQModel,
    "exaone4": Exaone4QModel,
    "grinmoe": GrinMoeQModel,
    "mllama": MLlamaQModel,
    "marin": Qwen3QModel,
    "granite": LlamaQModel, # 100% llama clone
    "granitemoehybrid": GraniteMoeHybridQModel,
    "mobilellm": MobileLLMQModel,
    "hymba": HymbaQModel,
    "olmo2": LlamaQModel, # 100% llama clone
    "ovis": OvisQModel,
    "ovis2": Ovis2QModel,
    "telechat": TeleChat2QModel,
    "instella": InstellaQModel,
    "mimo": MimoQModel,
    "falcon_h1": FalconH1QModel,
    "gpt_pangu": PanguAlphaQModel,
    "ernie4_5": Ernie4_5QModel,
    "ernie4_5_moe": Ernie4_5_MoeQModel,
    "seed_oss": LlamaQModel, # 100% llama clone
    "gpt_oss": GPTOSSGPTQ,
    "longcat_flash": LongCatFlashQModel,
    "llava_qwen2": LlavaQwen2QModel,
    "nemotron_h": NemotronHQModel,
    "bailing_moe": BailingMoeQModel,
    "lfm2_moe": LFM2MoeQModel,
    "llada2_moe": LLaDA2MoeQModel,
    "mistral3": Mistral3GPTQ,
    "afmoe": AfMoeQModel,
    "voxtral": VoxtralGPTQ,
}

if Qwen3_5QModel is not None:
    MODEL_MAP["qwen3_5"] = Qwen3_5QModel
    MODEL_MAP["qwen3_5_text"] = Qwen3_5QModel

if Qwen3_5_MoeQModel is not None:
    MODEL_MAP["qwen3_5_moe"] = Qwen3_5_MoeQModel

SUPPORTED_MODELS = list(MODEL_MAP.keys())


def _is_supported_quantization_config(config: AutoConfig) -> bool:
    quantization_config = getattr(config, "quantization_config", None)
    if not isinstance(quantization_config, dict):
        return False

    quant_format = quantization_config.get("quant_format")
    if isinstance(quant_format, str) and quant_format.lower() in (
        METHOD.GPTQ,
        METHOD.AWQ,
        METHOD.QQQ,
    ):
        return True

    quant_method = quantization_config.get("quant_method")
    if isinstance(quant_method, str) and quant_method.lower() in (
        METHOD.GPTQ,
        METHOD.AWQ,
        METHOD.QQQ,
    ):
        return True

    return False


def check_and_get_model_definition(model_dir, trust_remote_code=False):
    trust_remote_code = resolve_trust_remote_code(model_dir, trust_remote_code=trust_remote_code)
    config = AutoConfig.from_pretrained(model_dir, trust_remote_code=trust_remote_code)
    model_type = config.model_type.lower()

    # if model_type is not supported, use BaseQModel, will use auto_detect_module_tree to generate module tree
    if model_type not in SUPPORTED_MODELS:
        return BaseQModel

    return MODEL_MAP[model_type]

class GPTQModel:
    def __init__(self):
        raise EnvironmentError(
            "GPTQModel is not designed to be instantiated\n"
            "use `GPTQModel.from_pretrained` to load pretrained model and prepare for quantization via `.quantize()`.\n"
            "use `GPTQModel.from_quantized` to inference with post-quantized model."
        )

    @classmethod
    def load(
            cls,
            model_id_or_path: Optional[str],
            quantize_config: Optional[QuantizeConfig | Dict] = None,
            device_map: Optional[Union[str, Dict[str, Union[str, int]]]] = None,
            device: Optional[Union[str, torch.device]] = None,
            backend: Union[str, BACKEND] = BACKEND.AUTO,
            trust_remote_code: bool = False,
            **kwargs,
    ):
        if isinstance(model_id_or_path, str):
            model_id_or_path = model_id_or_path.strip()
        requested_trust_remote_code = trust_remote_code
        trust_remote_code = resolve_trust_remote_code(model_id_or_path, trust_remote_code=trust_remote_code)

        # normalize config to cfg instance
        if isinstance(quantize_config, Dict):
            quantize_config = QuantizeConfig(**quantize_config)

        if isinstance(backend, str):
            backend = BACKEND(backend)

        is_gptqmodel_quantized = False
        treat_as_local_path = isinstance(model_id_or_path, str) and (
            isdir(model_id_or_path) or os.path.isabs(model_id_or_path)
        )

        model_cfg = None
        if not (treat_as_local_path and not isdir(model_id_or_path)):
            model_cfg = AutoConfig.from_pretrained(model_id_or_path, trust_remote_code=trust_remote_code)

        if model_cfg is not None and _is_supported_quantization_config(model_cfg):
            # only if the model is quantized or compatible with gptqmodel should we set is_quantized to true
            is_gptqmodel_quantized = True
        else:
            # TODO FIX ME...not decoded to check if quant method is compatible or quantized by gptqmodel
            for name in [QUANT_CONFIG_FILENAME, "quant_config.json"]:
                if treat_as_local_path:  # Local paths should never trigger remote Hub lookups
                    if isdir(model_id_or_path) and os.path.exists(join(model_id_or_path, name)):
                        is_gptqmodel_quantized = True
                        break
                else:  # Remote
                    files = list_repo_files(repo_id=model_id_or_path)
                    for f in files:
                        if f == name:
                            is_gptqmodel_quantized = True
                            break

        if is_gptqmodel_quantized:
            m = cls.from_quantized(
                model_id_or_path=model_id_or_path,
                device_map=device_map,
                device=device,
                backend=backend,
                trust_remote_code=trust_remote_code,
                tokenizer_trust_remote_code=requested_trust_remote_code,
                **kwargs,
            )
        else:
            m = cls.from_pretrained(
                model_id_or_path=model_id_or_path,
                quantize_config=quantize_config,
                device_map=device_map,
                device=device,
                trust_remote_code=trust_remote_code,
                tokenizer_trust_remote_code=requested_trust_remote_code,
                **kwargs,
            )

        # debug model structure
        # if debug:
        #     print_module_tree(m.model)

        return m


    @classmethod
    def from_pretrained(
            cls,
            model_id_or_path: str,
            quantize_config: QuantizeConfig,
            trust_remote_code: bool = False,
            **model_init_kwargs,
    ) -> BaseQModel:
        requested_trust_remote_code = trust_remote_code
        tokenizer_trust_remote_code = model_init_kwargs.pop(
            "tokenizer_trust_remote_code",
            requested_trust_remote_code,
        )
        trust_remote_code = resolve_trust_remote_code(model_id_or_path, trust_remote_code=trust_remote_code)
        config = AutoConfig.from_pretrained(model_id_or_path, trust_remote_code=trust_remote_code)
        if _is_supported_quantization_config(config):
            log.warn("Model is already quantized, will use `from_quantized` to load quantized model.\n"
                           "If you want to quantize the model, please pass un_quantized model path or id, and use "
                           "`from_pretrained` with `quantize_config`.")
            return cls.from_quantized(model_id_or_path, trust_remote_code=trust_remote_code)

        if quantize_config and quantize_config.dynamic:
            log.warn(
                "GPTQModel's per-module `dynamic` quantization feature is fully supported in latest vLLM and SGLang but not yet available in hf transformers.")

        model_definition = check_and_get_model_definition(model_id_or_path, trust_remote_code)

        return model_definition.from_pretrained(
            pretrained_model_id_or_path=model_id_or_path,
            quantize_config=quantize_config,
            trust_remote_code=trust_remote_code,
            tokenizer_trust_remote_code=tokenizer_trust_remote_code,
            **model_init_kwargs,
        )

    @classmethod
    def from_quantized(
            cls,
            model_id_or_path: Optional[str],
            device_map: Optional[Union[str, Dict[str, Union[str, int]]]] = None,
            device: Optional[Union[str, int]] = None,
            backend: Union[str, BACKEND] = BACKEND.AUTO,
            adapter: Optional[Adapter | Dict] = None,
            trust_remote_code: bool = False,
            **kwargs,
    ) -> BaseQModel:
        requested_trust_remote_code = trust_remote_code
        tokenizer_trust_remote_code = kwargs.pop("tokenizer_trust_remote_code", requested_trust_remote_code)
        trust_remote_code = resolve_trust_remote_code(model_id_or_path, trust_remote_code=trust_remote_code)
        # normalize adapter to instance
        adapter = normalize_adapter(adapter)

        print(f"from_quantized: adapter: {adapter}")
        model_definition = check_and_get_model_definition(model_id_or_path, trust_remote_code)

        if isinstance(backend, str):
            backend = BACKEND(backend)

        return model_definition.from_quantized(
            model_id_or_path=model_id_or_path,
            device_map=device_map,
            device=device,
            backend=backend,
            trust_remote_code=trust_remote_code,
            tokenizer_trust_remote_code=tokenizer_trust_remote_code,
            adapter=adapter,
            **kwargs,
        )

    @classmethod
    def eval(
            cls,
            model_or_id_or_path: str=None,
            tokenizer: Union[PreTrainedTokenizerBase, Tokenicer]=None,
            tasks: Union[EVAL.LM_EVAL, EVAL.EVALPLUS, List[EVAL.LM_EVAL], List[EVAL.EVALPLUS], EVAL.MMLU_PRO, List[EVAL.MMLU_PRO]] = None, # set to None to fix mutable warning
            framework: Union[Type[EVAL.LM_EVAL],Type[EVAL.EVALPLUS],Type[EVAL.MMLU_PRO]] = EVAL.LM_EVAL,
            batch_size: Union[int, str] = 1,
            trust_remote_code: bool = False,
            output_path: Optional[str] = None,
            llm_backend: str = 'gptqmodel',
            backend: BACKEND = BACKEND.AUTO, # gptqmodel arg only
            random_seed: int = 1234,  # only for framework=EVAL.LM_EVAL backend=vllm
            model_args: Dict[str, Any] = None,  # only for framework=EVAL.LM_EVAL backend=vllm
            ntrain: int = 1,  # only for framework=EVAL.MMLUPRO
            **args
    ):
        if isinstance(model_or_id_or_path, str):
            trust_remote_code = resolve_trust_remote_code(model_or_id_or_path, trust_remote_code=trust_remote_code)

        from peft import PeftModel
        if model_args is None:
            model_args = {}
        if tasks is None:
            if framework == EVAL.LM_EVAL:
                tasks = [EVAL.LM_EVAL.ARC_CHALLENGE]
            elif framework == EVAL.MMLU_PRO:
                tasks = [EVAL.MMLU_PRO.MATH]
            else:
                tasks = [EVAL.EVALPLUS.HUMAN]

        elif not isinstance(tasks, List):
            tasks = [tasks]

        if framework is None:
            raise ValueError("Eval parameter: `framework` cannot be set to None")

        if not isinstance(tasks, list):
            raise ValueError("Eval parameter: `tasks` must be of List type")

        if llm_backend not in ['gptqmodel', 'vllm']:
            raise ValueError('Eval framework support llm_backend: [gptqmodel, vllm]')

        if llm_backend == "vllm":
            if "tensor_parallel_size" not in model_args:
                try:
                    cuda_devices = torch.cuda.device_count() if torch.cuda.is_available() else 0
                except Exception:
                    cuda_devices = 0
                if cuda_devices:
                    model_args["tensor_parallel_size"] = cuda_devices
            if "gpu_memory_utilization" not in model_args:
                model_args["gpu_memory_utilization"] = 0.90

        if isinstance(model_or_id_or_path, str):
            load_backend = backend
            if llm_backend == "vllm":
                disallowed_keys = {"pretrained", "tokenizer", "gptqmodel", "trust_remote_code", "backend", "model_id_or_path"}
                load_kwargs = {k: v for k, v in model_args.items() if k not in disallowed_keys}
            else:
                load_kwargs = model_args

            backend_name = load_backend.value if isinstance(load_backend, BACKEND) else str(load_backend)
            log.info(f"Eval: loading using backend = `{backend_name}`")
            model = GPTQModel.load(
                model_id_or_path=model_or_id_or_path,
                backend=load_backend,
                trust_remote_code=trust_remote_code,
                **load_kwargs,
            )
            model_id_or_path = model_or_id_or_path
        elif isinstance(model_or_id_or_path, BaseQModel) or isinstance(model_or_id_or_path, (PreTrainedModel, PeftModel)):
            model = model_or_id_or_path
            model_id_or_path = model.config.name_or_path  #
        else:
            raise ValueError(f"`model_or_id_or_path` is invalid. expected: `model instance or str` actual: `{model_or_id_or_path}`")

        if tokenizer is None:
            if isinstance(model, BaseQModel):
                tokenizer = model.tokenizer
            elif isinstance(model, PreTrainedModel) or model_id_or_path.strip():
                tokenizer = Tokenicer.load(model_id_or_path.strip())

        if tokenizer is None:
            raise ValueError("Tokenizer: Auto-loading of tokenizer failed with `model_or_id_or_path`. Please pass in `tokenizer` as argument.")

        if llm_backend == "gptqmodel": # vllm loads tokenizer
            model_args["tokenizer"] = tokenizer

        if framework == EVAL.LM_EVAL:
            from lm_eval.utils import make_table  # hack: circular import

            for task in tasks:
                if task not in EVAL.get_task_enums():
                    raise ValueError(f"Eval.lm_eval supported `tasks`: `{EVAL.get_all_tasks_string()}`, actual = `{task}`")

            model_name = "hf" if llm_backend == "gptqmodel" else llm_backend

            if llm_backend == "gptqmodel" and isinstance(model, BaseQModel) and model.quantized:
                model_args["gptqmodel"] = True
            model_args["pretrained"] = model_id_or_path

            # TODO FIXME lm-eval latest broken imports
            # lm_eval indirectly imports sglang, which expects newer Triton helpers.
            # Provide shims so import succeeds on older triton/runtime builds.
            try:
                from triton.runtime import cache as triton_cache

                if not hasattr(triton_cache, "default_cache_dir"):
                    triton_cache.default_cache_dir = lambda: getattr(triton_cache.knobs.cache, "dir", None)
                if not hasattr(triton_cache, "default_override_dir"):
                    triton_cache.default_override_dir = lambda: getattr(triton_cache.knobs.cache, "override_dir", None)
                if not hasattr(triton_cache, "default_dump_dir"):
                    triton_cache.default_dump_dir = lambda: getattr(triton_cache.knobs.cache, "dump_dir", None)
            except Exception as exc:
                log.warning("Triton cache shim failed; lm_eval import may fail: %s", exc)

            try:
                from lm_eval import simple_evaluate
                from lm_eval.models.huggingface import HFLM
            except BaseException as e:
                raise ValueError(f"lm_eval import failed: {e}. Please install via `pip install gptqmodel[eval]`.") from e

            if llm_backend == "gptqmodel" and model is not None:
                model_name = HFLM(
                    pretrained=model,
                    batch_size=batch_size,
                    trust_remote_code=trust_remote_code,
                )

            gen_kwargs = args.pop("gen_kwargs", None)

            # use model.generation_config whenever possible
            if gen_kwargs is None:
                # TODO: move to utils
                if hasattr(model, "generation_config") and isinstance(model.generation_config, GenerationConfig):
                    gen_dict = {
                        "do_sample": model.generation_config.do_sample,
                        "temperature": model.generation_config.temperature,
                        "top_k": model.generation_config.top_k,
                        "top_p": model.generation_config.top_p,
                        "min_p": model.generation_config.min_p,

                    }
                    gen_kwargs = ','.join(f"{key}={value}" for key, value in gen_dict.items() if value not in ["", {}, None, []])
                else:
                    gen_kwargs = "temperature=0.0,top_k=50" # default

            log.info(f"LM-EVAL: `gen_kwargs` = `{gen_kwargs}`")

            # lm-eval has very low scores if apply_chat_template is enabled
            apply_chat_template = args.pop("apply_chat_template", False) # args.pop("apply_chat_template", True if tokenizer.chat_template is not None else False)
            log.info(f"LM-EVAL: `apply_chat_template` = `{apply_chat_template}`")

            # TODO FIXME lm-eval latest broken imports
            # lm_eval pretty prints task yaml paths using Path.relative_to; when custom tasks live outside the
            # installed lm_eval package tree this raises ValueError. Monkeypatch to fall back to absolute paths.
            from pathlib import Path
            original_relative_to = Path.relative_to

            def _relative_to_noerror(self, other, *extra_args, **kwargs):
                try:
                    return original_relative_to(self, other, *extra_args, **kwargs)
                except ValueError:
                    return self

            Path.relative_to = _relative_to_noerror
            try:
                results = simple_evaluate(
                    model=model_name,
                    model_args=model_args,
                    tasks=[task.value for task in tasks],
                    batch_size=batch_size,
                    apply_chat_template=apply_chat_template,
                    gen_kwargs=gen_kwargs,
                    random_seed=random_seed,
                    numpy_random_seed=random_seed,
                    torch_random_seed=random_seed,
                    fewshot_random_seed=random_seed,
                    **args,
                )
            finally:
                Path.relative_to = original_relative_to

            if results is None:
                raise ValueError('lm_eval run fail, check your code!!!')

            print('--------lm_eval Eval Result---------')
            print(make_table(results))
            if "groups" in results:
                print(make_table(results, "groups"))
            print('--------lm_eval Result End---------')
            return results
        elif framework == EVAL.EVALPLUS:
            for task in tasks:
                if task not in EVAL.get_task_enums():
                    raise ValueError(f"evalplus support tasks: {EVAL.get_all_tasks_string()}")
            from ..utils.eval import evalplus, evalplus_make_table

            results = {}
            for task in tasks:
                base_formatted, plus_formatted, result_path = evalplus(
                    model=model_id_or_path,
                    dataset=task.value,
                    batch=batch_size,
                    trust_remote_code=trust_remote_code,
                    output_file=output_path,
                    backend=llm_backend
                )
                results[task.value] = {"base tests": base_formatted, "base + extra tests": plus_formatted,
                                       "results_path": result_path}
            print('--------evalplus Eval Result---------')
            evalplus_make_table(results)
            print('--------evalplus Result End---------')
            return results
        elif framework == EVAL.MMLU_PRO:
            for task in tasks:
                if task not in EVAL.get_task_enums():
                    raise ValueError(f"eval support tasks: {EVAL.get_all_tasks_string()}")
            from ..utils.mmlupro import mmlupro
            selected_subjects = ",".join(tasks)
            results = mmlupro(model,
                              tokenizer,
                              save_dir=output_path,
                              seed=random_seed,
                              selected_subjects=selected_subjects,
                              ntrain=ntrain,
                              batch_size=batch_size,
                              max_samples=args.pop("max_samples", None))

            print('--------MMLUPro Eval Result---------')
            print(results)
            print('--------MMLUPro Result End---------')
            return results
        else:
            raise ValueError("Eval framework support: EVAL.LM_EVAL, EVAL.EVALPLUS, EVAL.MMLUPRO")

    @staticmethod
    def export(model_id_or_path: str, target_path: str, format: str, trust_remote_code: bool = False):
        trust_remote_code = resolve_trust_remote_code(model_id_or_path, trust_remote_code=trust_remote_code)
        # load config
        config = AutoConfig.from_pretrained(model_id_or_path, trust_remote_code=trust_remote_code)

        if not config.quantization_config:
            raise ValueError("Model is not quantized")

        gptq_config = config.quantization_config

        # load gptq model
        gptq_model = GPTQModel.load(model_id_or_path, backend=BACKEND.TORCH)

        if format == "mlx":
            try:
                from mlx_lm.utils import save_config, save_model

                from ..utils.mlx import convert_gptq_to_mlx_weights
            except ImportError:
                raise ValueError(
                    "MLX not installed. Please install via `pip install gptqmodel[mlx] --no-build-isolation`.")

            mlx_weights, mlx_config = convert_gptq_to_mlx_weights(model_id_or_path, gptq_model, gptq_config,
                                                                  gptq_model.lm_head)

            save_model(target_path, mlx_weights, donate_model=True)

            save_config(mlx_config, config_path=target_path + "/config.json")
        elif format == "hf":
            from ..nn_modules.qlinear.torch import dequantize_model

            dequantized_model = dequantize_model(gptq_model.model)
            dequantized_model.save_pretrained(target_path)

        # save tokenizer to target path
        gptq_model.tokenizer.save_pretrained(target_path)

    # Use HfAPI and not Transformers to do upload
    @staticmethod
    def push_to_hub(repo_id: str,
                    quantized_path: str,  # saved local directory path
                    private: bool = False,
                    exists_ok: bool = False,  # set to true if repo already exists
                    token: Optional[str] = None,
                    ):

        if not quantized_path:
            raise RuntimeError("You must pass quantized model path as str to push_to_hub.")

        if not repo_id:
            raise RuntimeError("You must pass repo_id as str to push_to_hub.")

        from huggingface_hub import HfApi
        repo_type = "model"

        api = HfApi()
        # if repo does not exist, create it
        try:
            api.repo_info(repo_id=repo_id, repo_type=repo_type, token=token)
        except Exception:
            api.create_repo(repo_id=repo_id, repo_type=repo_type, token=token, private=private, exist_ok=exists_ok)

        # upload the quantized save folder
        api.upload_large_folder(
            folder_path=quantized_path,
            repo_id=repo_id,
            repo_type=repo_type,
        )

    class adapter:
        @classmethod
        def generate(
            cls,
            # eora adapter generation needs config Lora(rank=1, path='lora.safetensors')
            adapter: Adapter,
            model_id_or_path: str, # native model
            quantized_model_id_or_path: str, # gptqmodel quantized model
            calibration_dataset: Union[List[Dict[str, Union[List[int], torch.LongTensor]]], List[str], List[int]],
            calibration_dataset_concat_size: Optional[int] = None,
            calibration_dataset_sort: Optional[str] = None,
            batch_size: Optional[int] = 1,
            tokenizer: Optional[PreTrainedTokenizerBase] = None,
            calibration_concat_separator: Optional[str] = None,
            # pass-through vars for load()
            trust_remote_code: bool = False,
            dtype: Optional[Union[str, torch.dtype]] = None,
        ):
            if not adapter or not isinstance(adapter, Lora):
                raise ValueError(f"Adapter: expected `adapter` type to be `Lora`: actual = `{adapter}`.")

            adapter.validate_path(local=True)

            log.info("Model: Quant Model Loading...")
            quantized_model = GPTQModel.load(
                model_id_or_path=quantized_model_id_or_path,
                backend=BACKEND.TORCH,
                device=CPU,
                trust_remote_code=trust_remote_code,
                dtype=dtype,
            )

            qcfg = quantized_model.quantize_config
            qModules: Dict[str, TorchQuantLinear] = find_modules(module=quantized_model.model, layers=[TorchQuantLinear])
            # for name, module in qModules.items():
            #     quantized_weights[name] = module.dequantize_weight()
            del quantized_model
            torch_empty_cache()

            log.info("Model: Native Model Loading...")
            model = GPTQModel.load(
                model_id_or_path=model_id_or_path,
                quantize_config=qcfg,
                backend=BACKEND.TORCH,
                trust_remote_code=trust_remote_code,
                dtype=dtype,
                device=CPU,
            )

            log.info("Model: Adapter generation started")
            model._eora_generate(
                adapter=adapter,
                quantized_modules=qModules,
                calibration_dataset=calibration_dataset,
                calibration_dataset_concat_size=calibration_dataset_concat_size,
                calibration_dataset_sort=calibration_dataset_sort,
                batch_size=batch_size,
                tokenizer=tokenizer,
                calibration_concat_separator=calibration_concat_separator,
            )
            return



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import torch
from torch import nn

from ...adapter.adapter import Adapter, Lora
from ...models._const import DEVICE, PLATFORM
from ...nn_modules.qlinear import AWQuantLinear
from ...quantization import FORMAT, METHOD
from ...quantization.awq.utils.module import try_import
from ...utils.backend import BACKEND
from ...utils.gemv import calculate_zeros_width
from ...utils.logger import setup_logger


log = setup_logger()

awq_ext, msg = try_import("gptqmodel_awq_kernels")

class AwqGEMVQuantLinear(AWQuantLinear):
    SUPPORTS_BACKENDS = [BACKEND.GEMV]
    SUPPORTS_METHODS = [METHOD.AWQ]
    SUPPORTS_FORMATS = {FORMAT.GEMV: 40}
    SUPPORTS_BITS = [4]
    SUPPORTS_GROUP_SIZE = [-1, 16, 32, 64, 128]
    SUPPORTS_DESC_ACT = [True, False]
    SUPPORTS_SYM = [True, False]
    SUPPORTS_SHARDS = True
    SUPPORTS_TRAINING = True
    SUPPORTS_AUTO_PADDING = False
    SUPPORTS_IN_FEATURES_DIVISIBLE_BY = [1]
    SUPPORTS_OUT_FEATURES_DIVISIBLE_BY = [1]

    SUPPORTS_DEVICES = [DEVICE.ALL]
    SUPPORTS_PLATFORM = [PLATFORM.ALL]
    SUPPORTS_PACK_DTYPES = [torch.int32]
    SUPPORTS_ADAPTERS = [Lora]

    SUPPORTS_DTYPES = [torch.float16]

    # for transformers/optimum tests compat
    QUANT_TYPE = "awq_gemv"

    def __init__(
        self,
        bits: int,
        group_size: int,
        sym: bool,
        desc_act: bool,
        in_features: int,
        out_features: int,
        bias: bool = False,
        pack_dtype: torch.dtype = torch.int32,
        adapter: Adapter = None,
        register_buffers: bool = False,
        **kwargs,
    ):
        backend = kwargs.pop("backend", BACKEND.GEMV)
        super().__init__(
            bits=bits,
            group_size=group_size,
            sym=sym,
            desc_act=desc_act,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            pack_dtype=pack_dtype,
            backend=backend,
            adapter=adapter,
            register_buffers=False,
            **kwargs)

        self.split_k_iters = 8

        self.bias = None

        if register_buffers:
            self.register_buffer(
                "qweight",
                torch.zeros((out_features, in_features // self.pack_factor), dtype=self.pack_dtype),
            )
            self.register_buffer(
                "qzeros",
                torch.zeros(
                    out_features,
                    calculate_zeros_width(in_features, self.group_size),
                    dtype=self.pack_dtype,
                ),
            )
            self.register_buffer(
                "scales",
                torch.zeros(
                    out_features,
                    calculate_zeros_width(in_features, self.group_size) * self.pack_factor,
                    dtype=torch.float16,
                ),
            )

            if bias:
                self.register_buffer("bias", torch.zeros(out_features, dtype=torch.float16))

    def post_init(self):
        # if self.padded_infeatures != self.in_features:
        #     self.qweight.resize_(self.padded_infeatures // self.pack_dtype_bits * self.bits, self.out_features)
        #     self.qzeros.resize_(
        #         math.ceil(self.padded_infeatures / self.group_size),
        #         self.out_features // self.pack_dtype_bits * self.bits
        #     )
        #     self.scales.resize_((math.ceil(self.padded_infeatures / self.group_size), self.out_features), )
        #     self.g_idx = torch.tensor([i // self.group_size for i in range(self.padded_infeatures)], dtype=torch.int32,
        #                               device=self.g_idx.device)

        # awq only accepts float16
        self.scales = self.scales.to(dtype=torch.float16)

        super().post_init()

    def forward(self, x: torch.Tensor):
        if awq_ext is None:
            raise ModuleNotFoundError("External AWQ kernels are not properly installed." + msg)

        out_shape = x.shape[:-1] + (self.out_features,)
        inputs = x.reshape(-1, x.shape[-1])

        input_dtype = inputs.dtype
        if input_dtype != torch.float16:
            inputs = inputs.half()

        if inputs.shape[0] > 8:
            out = awq_ext.gemmv2_forward_cuda(
                inputs,
                self.qweight,
                self.scales,
                self.qzeros,
                self.group_size,
                self.split_k_iters,
            )
        else:
            out = awq_ext.gemv_forward_cuda(
                inputs, self.qweight, self.scales, self.qzeros, self.group_size
            )

        if input_dtype != torch.float16:
            out = out.to(dtype=input_dtype)

        out = out + self.bias if self.bias is not None else out

        if self.adapter:
            out = self.adapter.apply(x=x, out=out)

        return out.reshape(out_shape)

    def pack(self, linear: nn.Module, scales: torch.Tensor, zeros: torch.Tensor, g_idx: torch.Tensor=None):
        # need scales and zeros info for real quantization
        assert scales is not None and zeros is not None
        scale_zeros = zeros * scales

        pack_num = 32 // self.bits
        qscales = torch.zeros(
            (
                scales.shape[0],
                calculate_zeros_width(linear.in_features, self.group_size) * pack_num,
            ),
            dtype=torch.float16,
            device=scales.device,
        )
        qscales[:, : scales.shape[1]] = scales
        self.register_buffer("scales", qscales)
        if linear.bias is not None:
            self.register_buffer("bias", linear.bias.clone().half())
        else:
            self.bias = None

        intweight = []
        for idx in range(self.in_features):
            intweight.append(
                torch.round(
                    (linear.weight.data[:, idx] + scale_zeros[:, idx // self.group_size])
                    / self.scales[:, idx // self.group_size]
                ).to(torch.int)[:, None]
            )
        intweight = torch.cat(intweight, dim=1)
        intweight = intweight.to(dtype=torch.int32)
        qweight = torch.zeros(
            (intweight.shape[0], intweight.shape[1] // 32 * self.bits),
            dtype=torch.int32,
            device=intweight.device,
        )

        for col in range(intweight.shape[1] // pack_num):
            if self.bits == 4:
                order_map = [0, 1, 2, 3, 4, 5, 6, 7]
            else:
                raise NotImplementedError("Only 4-bit are supported for now.")
            for i in range(pack_num):
                qweight_col = intweight[:, col * pack_num + order_map[i]]
                qweight[:, col] |= qweight_col << (i * self.bits)
        self.register_buffer("qweight", qweight)

        zeros = zeros.to(dtype=torch.int32)
        qzeros = torch.zeros(
            (zeros.shape[0], calculate_zeros_width(linear.in_features, self.group_size)),
            dtype=torch.int32,
            device=zeros.device,
        )

        for col in range((zeros.shape[1] + pack_num - 1) // pack_num):
            if self.bits == 4:
                order_map = [0, 1, 2, 3, 4, 5, 6, 7]
            else:
                raise NotImplementedError("Only 4-bit are supported for now.")
            for i in range(pack_num):
                if col * pack_num + order_map[i] >= zeros.shape[1]:
                    continue
                qzero_col = zeros[:, col * pack_num + order_map[i]]
                qzeros[:, col] |= qzero_col << (i * self.bits)
        self.register_buffer("qzeros", qzeros)

    def extra_repr(self) -> str:
        return (
            "in_features={}, out_features={}, bias={}, bits={}, group_size={}".format(
                self.in_features,
                self.out_features,
                self.bias is not None,
                self.bits,
                self.group_size,
            )
        )

__all__ = ["AwqGEMVQuantLinear"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from ..base import BaseQModel


class CodeGenQModel(BaseQModel):
    pre_lm_head_norm_module = "transformer.ln_f"

    module_tree = [
        "transformer",
        "h",
        "#",
        {
            "attn": ("qkv_proj:0", "out_proj:1"),
            "mlp": ("fc_in:0", "fc_out:1"),
        }
    ]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

"""Simple pause/resume functionality for GPTQModel quantization.

Provides thread-safe pause/resume capabilities with keyboard input handling.
"""

import atexit
import logging
import sys
import threading
import time
from contextlib import contextmanager
from enum import Enum
from typing import Callable, Dict, List, Optional

from .colors import ANSIColor, color_text


log = logging.getLogger(__name__)

_IS_WINDOWS = sys.platform == "win32"

if _IS_WINDOWS:
    import msvcrt
else:
    try:
        import select
        import termios
        import tty
    except ImportError:
        log.warning("termios, tty, or select not available. Pause/resume from keyboard will not work.")


_original_termios_settings = None
_terminal_restored = False
_terminal_restore_lock = threading.Lock()


def _restore_terminal_settings_on_exit():
    """Restore terminal settings on exit. Idempotent, thread-safe, and safe to call multiple times."""
    global _original_termios_settings, _terminal_restored

    with _terminal_restore_lock:
        # Idempotent: skip if already restored or nothing to restore
        if _terminal_restored or _original_termios_settings is None:
            return

        try:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, _original_termios_settings)
            # Show cursor explicitly (in case it was hidden by progress bars)
            # ANSI escape sequence DECSET: Show cursor
            sys.stdout.write('\x1b[?25h')
            _original_termios_settings = None
            _terminal_restored = True
            log.debug("Terminal settings restored on exit.")
        except Exception as e:
            log.warning(f"Failed to restore terminal settings on exit: {e}")
            # Mark as restored even on failure to prevent retry loops
            _terminal_restored = True


if not _IS_WINDOWS and 'termios' in sys.modules:
    if sys.stdin.isatty():
        atexit.register(_restore_terminal_settings_on_exit)


class PauseResumeState(Enum):
    """Enumeration of possible pause/resume states."""
    RUNNING = "running"
    PAUSE_REQUESTED = "pause_requested"
    PAUSED = "paused"


class PauseResumeController:
    """
    Simple thread-safe controller for managing pause/resume during quantization.

    Provides:
    - Thread-safe state management
    - Keyboard input handling from stdin ('p' key)
    - Integration with progress tracking
    """

    def __init__(self):
        """
        Initialize the pause/resume controller.
        """
        self._state = PauseResumeState.RUNNING
        self._state_lock = threading.RLock()
        self._pause_event = threading.Event()
        self._resume_event = threading.Event()

        # Stdin input handling
        self._input_thread = None
        self._stop_event = threading.Event()

        # Callbacks for status updates
        self._status_callback: Optional[Callable[[PauseResumeState], None]] = None

        # Progress bar references for immediate title updates
        self._progress_bars: List[Dict] = []  # List of {"pb": progress_bar, "title_func": callable}

        # Initialize events
        self._resume_event.set()  # Allow execution to start

    def _play_icon(self) -> str:
        return color_text(">", ANSIColor.GREEN)

    def _pause_icon(self) -> str:
        return color_text("||", ANSIColor.YELLOW)

    def get_status_hint(self) -> str:
        """Get status hint for main progress bar."""
        state = self.get_state()
        if state == PauseResumeState.RUNNING:
            return f"['p' to {self._pause_icon()}]"
        elif state == PauseResumeState.PAUSE_REQUESTED:
            return f"[will {self._pause_icon()} at layer end]"
        elif state == PauseResumeState.PAUSED:
            return f"['p' to {self._play_icon()}]"
        else:
            return ""

    def status_icon(self) -> str:
        """Get current status icon (|| for pause states, > for running)."""
        state = self.get_state()
        if state in [PauseResumeState.RUNNING, PauseResumeState.PAUSE_REQUESTED]:
            return self._play_icon()
        elif state in [PauseResumeState.PAUSED]:
            return self._pause_icon()
        else:
            return ""

    def wrap_text(self, text: str) -> str:
        """Wrap text with status icon and hint for progress bar display."""
        icon = self.status_icon()
        hint = self.get_status_hint()

        if icon:
            text = f"{icon} {text}"
        if hint:
            text = f"{text} {hint}"
        return text

    def register_and_draw_progress_bar(self, pb, title: Optional[str] = None, subtitle: Optional[str] = None):
        """
        Register a progress bar for immediate title updates when pause/resume state changes.

        Args:
            pb: Progress bar instance
            title: Optional title string without status icons/hints
            subtitle: Optional subtitle string
        """
        # update progress bar subtitle
        pb.subtitle(subtitle)
        # register or update base title for progress bar
        try:
            with self._state_lock:
                # Check if this progress bar is already registered
                for item in self._progress_bars:
                    if item["pb"] == pb:
                        # Update the title and subtitle if already registered
                        item["title"] = title
                        return

                # Register new progress bar
                self._progress_bars.append({"pb": pb, "title": title})
        finally:
            self._update_progress_bars()

    def unregister_progress_bar(self, pb):
        """
        Unregister a progress bar from immediate title updates.

        Args:
            pb: Progress bar instance to remove
        """
        with self._state_lock:
            self._progress_bars = [item for item in self._progress_bars if item["pb"] != pb]

    def _update_progress_bars(self):
        """Update all registered progress bars with current pause/resume status."""
        for item in self._progress_bars:
            pb = item["pb"]
            title = item.get("title")

            if title:
                wrapped_title = self.wrap_text(title)
                try:
                    pb.title(wrapped_title)
                    pb.draw()
                except Exception as e:
                    log.warning(f"Failed to update progress bar title: {e}")

    def _windows_stdin_listener(self):
        while not self._stop_event.is_set():
            try:
                if msvcrt.kbhit():
                    char = msvcrt.getch().decode('utf-8').lower()
                    if char == 'p':
                        self.toggle_pause_resume()
                time.sleep(0.1)
            except Exception as e:
                log.warning(f"Error in stdin listener: {e}")
                break

    def _posix_stdin_listener(self):
        global _original_termios_settings
        old_settings = None
        try:
            try:
                old_settings = termios.tcgetattr(sys.stdin)
            except Exception as e:
                # Some stdin providers used by tests and wrappers expose a stream that
                # reports as a TTY but does not support termios inspection.
                log.debug(f"Unable to read terminal settings from stdin: {e}")

            if old_settings is not None:
                if _original_termios_settings is None:
                    _original_termios_settings = old_settings
                try:
                    tty.setcbreak(sys.stdin.fileno())
                except Exception as e:
                    # Test harnesses and redirected stdin can expose a file descriptor that
                    # passes isatty() checks but still rejects termios mode changes.
                    log.debug(f"Unable to switch stdin to cbreak mode: {e}")

            while not self._stop_event.is_set():
                try:
                    ready, _, _ = select.select([sys.stdin], [], [], 0.1)
                except Exception as e:
                    log.debug(f"stdin polling unavailable for pause/resume listener: {e}")
                    time.sleep(0.1)
                    continue

                if ready:
                    try:
                        char = sys.stdin.read(1)
                    except Exception as e:
                        log.debug(f"stdin read unavailable for pause/resume listener: {e}")
                        time.sleep(0.1)
                        continue

                    if char.lower() == 'p':
                        self.toggle_pause_resume()
                        time.sleep(0.1)
        except Exception as e:
            log.warning(f"Error in stdin listener: {e}")
        finally:
            # Defense-in-depth: restore terminal settings when listener thread exits
            # This ensures cleanup even if cleanup() is never called
            if old_settings:
                _restore_terminal_settings_on_exit()

    def _setup_keyboard_handler(self):
        """Setup keyboard event handlers for pause/break keys."""
        if not sys.stdin.isatty():
            log.info("Not a TTY. Pause/resume from keyboard is disabled.")
            return

        target = None
        if _IS_WINDOWS:
            target = self._windows_stdin_listener
        else:
            if 'termios' in sys.modules:
                target = self._posix_stdin_listener
            else:
                log.warning("termios module not found, cannot listen for keyboard input on POSIX.")
                return

        self._input_thread = threading.Thread(target=target, daemon=True)
        self._input_thread.start()

    def set_status_callback(self, callback: Callable[[PauseResumeState], None]):
        """Set callback for state changes."""
        with self._state_lock:
            self._status_callback = callback

    def _set_state(self, new_state: PauseResumeState):
        """Internal method to change state with notification."""
        with self._state_lock:
            old_state = self._state
            self._state = new_state

            # Update events based on state
            if new_state == PauseResumeState.PAUSED:
                self._pause_event.set()
                self._resume_event.clear()
            elif new_state == PauseResumeState.RUNNING:
                self._pause_event.clear()
                self._resume_event.set()

            # Update progress bars immediately when state changes
            if old_state != new_state:
                self._update_progress_bars()

            # Notify callback
            if self._status_callback and old_state != new_state:
                try:
                    self._status_callback(new_state)
                except Exception as e:
                    log.warning(f"Status callback error: {e}")

    def get_state(self) -> PauseResumeState:
        """Get current state."""
        with self._state_lock:
            return self._state

    def pause(self):
        """Request pause - will pause at next safe point."""
        with self._state_lock:
            if self._state == PauseResumeState.RUNNING:
                self._set_state(PauseResumeState.PAUSE_REQUESTED)

    def resume(self):
        """Resume quantization."""
        with self._state_lock:
            if self._state == PauseResumeState.PAUSED or self._state == PauseResumeState.PAUSE_REQUESTED:
                self._set_state(PauseResumeState.RUNNING)

    def toggle_pause_resume(self):
        """Toggle between pause and resume states."""
        current_state = self.get_state()
        if current_state == PauseResumeState.RUNNING:
            self.pause()
        elif current_state in [PauseResumeState.PAUSED, PauseResumeState.PAUSE_REQUESTED]:
            self.resume()

    def check_pause_point(self, layer_info: Optional[str] = None) -> bool:
        """
        Check if we should pause at this point.

        Called between layers or other safe points.

        Args:
            layer_info: Optional description of current layer for logging

        Returns:
            Always True (execution continues)
        """
        # Check if pause was requested
        with self._state_lock:
            if self._state == PauseResumeState.PAUSE_REQUESTED:
                self._set_state(PauseResumeState.PAUSED)

        # Wait if paused
        if self._pause_event.is_set():
            while self._pause_event.is_set():
                if self._resume_event.wait(timeout=0.1):
                    with self._state_lock:
                        if self._state == PauseResumeState.PAUSED:
                            self._set_state(PauseResumeState.RUNNING)
                        break

        return True

    def is_paused(self) -> bool:
        """Check if currently paused or pause requested."""
        with self._state_lock:
            return self._state in [PauseResumeState.PAUSED, PauseResumeState.PAUSE_REQUESTED]

    def is_running(self) -> bool:
        """Check if currently running."""
        with self._state_lock:
            return self._state == PauseResumeState.RUNNING

    @contextmanager
    def pause_context(self, operation_name: str):
        """
        Context manager for operations that should respect pause state.

        Args:
            operation_name: Name of the operation for logging
        """
        self.check_pause_point(operation_name)
        try:
            yield
        finally:
            # Check pause point again after operation
            self.check_pause_point(f"after {operation_name}")

    @contextmanager
    def lifecycle(self):
        """
        Context manager for the controller lifecycle.

        Usage:
            with controller.lifecycle():
                # quantization code here
                pass
            # cleanup is guaranteed here
        """
        self._setup_keyboard_handler()

        try:
            yield
        finally:
            # This ALWAYS runs, even on exceptions or KeyboardInterrupt
            self.cleanup()

    def cleanup(self):
        """
        Cleanup resources and keyboard handlers.

        This method is idempotent - safe to call multiple times.
        Ensures terminal settings are restored on all exit paths.
        """
        # Stop the input listener thread
        self._stop_event.set()
        if self._input_thread:
            self._input_thread.join(timeout=0.5)
            self._input_thread = None

        # Clear progress bar references
        with self._state_lock:
            self._progress_bars.clear()

        # Set final state to running for clean shutdown
        with self._state_lock:
            if self._state != PauseResumeState.RUNNING:
                self._set_state(PauseResumeState.RUNNING)

        # Ensure terminal settings are restored (idempotent)
        if not _IS_WINDOWS and 'termios' in sys.modules:
            _restore_terminal_settings_on_exit()



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

"""Layer execution stage extracted from ModuleLooper."""

from __future__ import annotations

import logging
import threading
import time
from concurrent.futures import as_completed
from typing import TYPE_CHECKING, Dict, List, Optional

from defuser.modeling.replace_modules import materialize_model
from ..nn_modules.hooked_linear import replace_module_with_hooked_legacy
from ..nn_modules.converter import MODULE_CONVERTER_MAP
from ..quantization.config import GcMode
import torch

from .. import DEBUG_ON, DEVICE_THREAD_POOL
from ..looper.awq_processor import AWQProcessor
from ..looper.gptq_processor import GPTQProcessor
from ..looper.named_module import NamedModule
from ..looper.qqq_processor import QQQProcessor
from ..utils.device import get_device, get_device_new
from ..utils.logger import log_time_block, setup_logger
from ..utils.model import find_modules, get_module
from ..utils.offload import offload_to_disk
from ..utils.torch import CPU, torch_empty_cache, torch_sync
from .stage_subset import SubsetForwardContext, run_subset_stage

if TYPE_CHECKING:  # pragma: no cover - type hints only
    from .module_looper import ModuleLooper


def _find_last_quantized_layer_index(
    looper: "ModuleLooper",
    *,
    layer_modules: List[List[str]],
    layers_prefix: Optional[str],
    layer_count: int,
) -> Optional[int]:
    """Return the highest layer index whose tracked modules are not all dynamically skipped."""
    if looper.gptq_model.quantize_config.lm_head or not layers_prefix:
        return None

    layer_module_names = {
        name.split("#", 1)[0]
        for module_group in layer_modules
        for name in module_group
        if name
    }
    if not layer_module_names:
        return None

    last_quantized_layer_index = -1
    for candidate_layer_index in range(layer_count):
        for module_name in layer_module_names:
            module_full_name = f"{layers_prefix}.{candidate_layer_index}.{module_name}"
            # If at least one module in this layer is not dynamically excluded,
            # the layer still needs forward/quantization work.
            if looper.gptq_model.quantize_config.dynamic_get(layer_name=module_full_name) != False:
                last_quantized_layer_index = candidate_layer_index
                break

    return last_quantized_layer_index


def run_layer_stage(
    looper: 'ModuleLooper',
    *,
    layers: List[torch.nn.Module],
    layer_modules: List[List[str]],
    layers_prefix: Optional[str],
    failsafe,
    shared_kv_cache_dict: Dict[int, torch.Tensor],
    pb,
    layer_count: int,
    region_timer,
    finalize_progress_cls,
    logger=None,
) -> None:
    """Execute the main per-layer quantization loop."""
    # Trailing layers whose tracked modules are all dynamically excluded never
    # need another forward or finalize pass, so the loop can stop once the
    # final eligible layer has been processed.
    last_quantized_layer_index = _find_last_quantized_layer_index(
        looper,
        layer_modules=layer_modules,
        layers_prefix=layers_prefix,
        layer_count=layer_count,
    )

    log = logger or setup_logger()
    for layer_index in pb:
        # Iterate over every transformer layer (plus lm_head when enabled) as
        # progress-bar controlled units of work.
        if looper._check_loop_stop():
            break
        is_lm_head_module = layer_index >= layer_count

        if (
            not is_lm_head_module
            and last_quantized_layer_index is not None
            and layer_index > last_quantized_layer_index
        ):
            # The remaining layers are fully skipped by dynamic config, so
            # avoid entering another layer-level quantization cycle.
            log.debug(
                "StageLayer: early stop at layer=%s, last_quantized_layer=%s",
                layer_index,
                last_quantized_layer_index,
            )
            pb.close()
            break

        if is_lm_head_module:
            layer_title = "Quantizing lm_head"
            module = get_module(looper.gptq_model.model, key=looper.gptq_model.lm_head)
        else:
            layer_title = f"Quantizing layer {layer_index} of {layer_count - 1}"
            module = layers[layer_index]

        looper.pause_controller.register_and_draw_progress_bar(pb, title=layer_title, subtitle="")

        if module.__class__.__name__.lower() == "MllamaCrossAttentionDecoderLayer".lower():
            # TODO FIXME: currently we not support quantizing cross attention layer (pixel_values)
            continue

        module = looper.gptq_model.pre_quantize(module)

        if is_lm_head_module:
            layer_descriptor = looper.gptq_model.lm_head
        else:
            model_type = looper.gptq_model.model.config.model_type
            if model_type in MODULE_CONVERTER_MAP:
                converter = MODULE_CONVERTER_MAP[model_type]
                module = converter(module, looper.gptq_model.model.config)

            replace_module_with_hooked_legacy(module, quant_lm_head=looper.gptq_model.quantize_config.lm_head)

            layers[layer_index] = module

            if layers_prefix:
                layer_descriptor = f"{layers_prefix}.{layer_index}"
            else:
                layer_descriptor = str(layer_index)

        materialize_model(module)

        cur_layer_device = get_device(module)
        full = find_modules(module, name=looper.gptq_model.lm_head if is_lm_head_module else "")

        for p_index, processor in enumerate(looper.processors):
            # Each processor contributes a quantization phase; walk them in
            # order so their caches and side effects line up with the pipeline.
            processor.log_call_count = 0  # reset
            processor.collect_memory_info(layer_index)

            modules = [[looper.gptq_model.lm_head]] if is_lm_head_module else layer_modules

            # for NativeProcessor we process one time forward on all grouped module subsets
            if processor.fwd_all_modules_in_single_pass:
                # merge all subsets into one
                modules = [sum(modules, [])]

            layer_inputs = processor.inputs_cache.layer_inputs
            if is_lm_head_module and layer_inputs:
                layer_inputs = looper.gptq_model.lm_head_pre_quantize_generate_hook(layer_inputs)
            layer_input_kwargs = processor.inputs_cache.layer_input_kwargs
            position_ids = processor.inputs_cache.position_ids
            attention_masks = processor.inputs_cache.attention_masks

            processed_subset: Dict[str, NamedModule] = {}
            last_subset_context: Optional[SubsetForwardContext] = None
            previous_subset_processed: Optional[Dict[str, NamedModule]] = None

            subsets = []
            for names in modules:
                subset = looper.create_named_modules(
                    module=module,
                    full=full,
                    is_lm_head_module=is_lm_head_module,
                    layer_index=layer_index,
                    layers_prefix=layers_prefix,
                    names=names,
                    processor=processor,
                    failsafe=failsafe,
                    layer_module=module,
                )
                # Skip empty subsets caused by per-layer structure differences or dynamic config exclusions;
                # otherwise awq_processor may fail to quantize
                if subset:
                    subsets.append(subset)

            subset_total = len(subsets)
            for index, subset in enumerate(subsets):
                # Process the layer in smaller subsets so attention groups or
                # MoE experts can be quantized independently within a layer.
                if DEBUG_ON and log.isEnabledFor(logging.DEBUG):
                    if isinstance(processor, AWQProcessor):
                        log.debug(
                            "StageLayer[awq]: layer=%s subset=%s/%s size=%s names=%s",
                            layer_index,
                            index + 1,
                            subset_total,
                            len(subset),
                            subset[:5],
                        )
                    else:
                        log.debug(
                            "StageLayer: layer=%s subset=%s/%s processor=%s size=%s names=%s",
                            layer_index,
                            index + 1,
                            subset_total,
                            processor.name(),
                            len(subset),
                            subset[:8],
                        )
                subset_result = run_subset_stage(
                    looper=looper,
                    processor=processor,
                    module=module,
                    layer_inputs=layer_inputs,
                    layer_input_kwargs=layer_input_kwargs,
                    position_ids=position_ids,
                    attention_masks=attention_masks,
                    cur_layer_device=cur_layer_device,
                    is_lm_head_module=is_lm_head_module,
                    layer_descriptor=layer_descriptor,
                    layer_title=layer_title,
                    layer_index=layer_index,
                    layers_prefix=layers_prefix,
                    subset=subset,
                    subset_index=index,
                    subset_total=subset_total,
                    full=full,
                    failsafe=failsafe,
                    shared_kv_cache_dict=shared_kv_cache_dict,
                    pb=pb,
                    log=log,
                    region_timer=region_timer,
                    previous_processed_subset=previous_subset_processed,
                    subset_event_cb=looper._subset_event_dispatch,
                )

                layer_inputs = subset_result.layer_inputs
                processed_subset.update(subset_result.processed_subset)
                previous_subset_processed = subset_result.processed_subset
                if subset_result.forward_context is not None:
                    last_subset_context = subset_result.forward_context

            is_last_module = layer_index == len(pb) - 1
            layer_outputs: List[List[torch.Tensor]] = []
            subset_context = last_subset_context
            forward_device_map = subset_context.forward_device_map if subset_context else {}
            subset_forward_serial = subset_context.subset_forward_serial if subset_context else False
            subset_reference_total = subset_context.subset_total if subset_context else subset_total
            subset_reference_index = subset_context.subset_index if subset_context else max(subset_total - 1, 0)
            subset_for_overrides = subset_context.subset if subset_context else {}
            preserve_devices = bool(forward_device_map)

            # second forward after process()
            if not is_last_module and processor.fwd_after_process and subset_context is not None:
                replay_batch_count = looper._resolve_batch_total(
                    getattr(processor, "num_batches", None),
                    layer_inputs,
                )
                replay_row_counts = list(looper._collect_row_counts(layer_inputs))
                if not replay_row_counts and replay_batch_count > 0:
                    replay_row_counts = [1] * replay_batch_count
                if len(replay_row_counts) > replay_batch_count:
                    replay_row_counts = replay_row_counts[:replay_batch_count]
                replay_total_rows = sum(replay_row_counts) if replay_row_counts else replay_batch_count
                replay_total_rows = max(replay_total_rows, 1)
                if len(replay_row_counts) < replay_batch_count:
                    replay_row_counts.extend([1] * (replay_batch_count - len(replay_row_counts)))
                replay_msg = (
                    "Forward replay "
                    f"(layer=`{layer_descriptor}`, batches={replay_batch_count}, rows={replay_total_rows})"
                )
                replay_pb = (
                    log.pb(range(replay_total_rows))
                       .manual()
                       .set(show_left_steps=False)
                )
                replay_pb.title(replay_msg).subtitle(
                    f"Forward replay Row 0/{replay_total_rows}"
                ).draw()
                # Forward replay shares the same VRAM spike; block until the pool drains first.
                # DEVICE_THREAD_POOL.wait()
                # try to cleanup recent objects before forward
                #timed_gc_collect(1)

                replay_start = time.perf_counter()
                replay_source = f"{layer_descriptor}:subset{subset_reference_index + 1}/{subset_reference_total}"

                replay_prev_devices: Dict[str, torch.device] = {}
                if forward_device_map:
                    replay_prev_devices = looper._apply_forward_device_overrides(
                        subset_for_overrides,
                        forward_device_map,
                        fallback_modules=full,
                    )

                # if log.isEnabledFor(logging.DEBUG):
                #     replay_snapshot = []
                #     for name, named_module in subset.items():
                #         target_device = getattr(named_module, "target_device", None)
                #         if target_device is None:
                #             try:
                #                 target_device = get_device(named_module.module)
                #             except Exception:
                #                 target_device = None
                #         target_device_str = str(target_device) if target_device is not None else "unknown"
                #         replay_snapshot.append(f"{name}:{target_device_str}")
                #     log.debug(
                #         "ModuleLooper: Forward replay device snapshot (layer=`%s`, subset=%d/%d, serial=%s) %s",
                #         layer_descriptor,
                #         index + 1,
                #         subset_total,
                #         subset_forward_serial,
                #         ", ".join(replay_snapshot),
                #     )

                try:
                    # Reset current subset for MoE lifecycle hooks as we do not need to collect activation at this stage,
                    # and need to collect only outputs produced by original forward
                    looper._current_subset = None
                    
                    layer_outputs = looper._run_forward_batches(
                        module=module,
                        processor=processor,
                        layer_inputs=layer_inputs,
                        layer_input_kwargs=layer_input_kwargs,
                        position_ids=position_ids,
                        attention_masks=attention_masks,
                        cur_layer_device=cur_layer_device,
                        is_lm_head_module=is_lm_head_module,
                        shared_kv_cache_dict=shared_kv_cache_dict,
                        layer_index=layer_index,
                        need_outputs=True,
                        reuse_kv=False,
                        progress_pb=replay_pb,
                        progress_title=replay_msg,
                        progress_stage="Forward replay",
                        progress_rows_per_batch=replay_row_counts,
                        progress_total_rows=replay_total_rows,
                        force_serial=subset_forward_serial,
                        preserve_module_devices=preserve_devices,
                    )
                finally:
                    if forward_device_map:
                        looper._restore_forward_device_overrides(
                            subset_for_overrides,
                            replay_prev_devices,
                            fallback_modules=full,
                        )
                    if replay_pb is not None:
                        replay_pb.close()
                if region_timer is not None:
                    region_timer.record(
                        "post_quant_forward",
                        time.perf_counter() - replay_start,
                        source=replay_source,
                    )

            # Finalize module after last processor
            if p_index == len(looper.processors) - 1:
                torch_sync()

                if not is_lm_head_module:
                    layers[layer_index] = looper.gptq_model.post_quantize(module)
                else:
                    looper.gptq_model.post_quantize(module)

                for finalized in processed_subset.values():
                    # Reset finalized modules to CPU to guarantee deterministic
                    # ownership before the next processor touches the layer.
                    if isinstance(finalized, NamedModule):
                        setattr(finalized, "target_device", CPU)
                        inner_module = getattr(finalized, "module", None)
                    else:
                        inner_module = finalized

                    if inner_module is not None and hasattr(inner_module, "target_device"):
                        setattr(inner_module, "target_device", CPU)

                if region_timer is not None:
                    region_timer.flush()

            if processor.fwd_after_process:
                processor.clear_cache_data()
                processor.receive_layer_inputs(layer_outputs)
                layer_inputs = processor.inputs_cache.layer_inputs
                looper.pause_controller.register_and_draw_progress_bar(pb, title=layer_title, subtitle="")

            if p_index == len(looper.processors) - 1:
                torch_sync()

                # Gather finalize tasks (can offload to disk); run them via the pool
                finalize_tasks = []

                for reverse_p in reversed(looper.processors):
                    # Collect finalize tasks in reverse to mirror the processor
                    # execution order and honor downstream dependencies.
                    for module in processed_subset.values():
                        actual_module = module.module if isinstance(module, NamedModule) else module

                        get_device_new(
                            actual_module,
                            recursive=True,
                            assert_mode=True,
                            expected=CPU,
                        )
                        with looper._quant_device_lock:
                            key = getattr(module, "full_name", getattr(module, "name", None))
                            if key is not None:
                                looper._module_device_map[key] = CPU

                        target_dev = CPU
                        module_label = getattr(module, "full_name", getattr(module, "name", ""))
                        layer_idx = getattr(module, "layer_index", None)
                        finalize_tasks.append((reverse_p, module, module_label, target_dev, layer_idx))

                finalize_count = len(finalize_tasks)
                finalize_futures = []
                finalize_pb = log.pb(range(finalize_count)).manual().set(show_left_steps=False)

                @torch.inference_mode()
                def _finalize_on_worker(process, module, idx, total, module_label, layer_idx):
                    resolved_label = module_label or getattr(module, "full_name", getattr(module, "name", ""))
                    start = time.perf_counter() if region_timer is not None else None
                    try:
                        with log_time_block(
                            "submodule_finalize",
                            logger=log,
                            module_name=resolved_label,
                        ):
                            process.submodule_finalize(module, looper.gptq_model)

                        # Disk offload (lifecycle TODO note preserved)
                        if isinstance(process, (GPTQProcessor, QQQProcessor, AWQProcessor)):
                            quant_config = getattr(looper.gptq_model, "quantize_config", None)
                            if quant_config and getattr(quant_config, "offload_to_disk", False):
                                offload_path = getattr(quant_config, "offload_to_disk_path", None)
                                if offload_path:
                                    module_full_name = getattr(module, "full_name", None)
                                    target_module = (
                                        looper.gptq_model.model.get_submodule(module_full_name)
                                        if module_full_name
                                        else module
                                    )
                                    offload_start = time.perf_counter() if region_timer is not None else None
                                    with log_time_block(
                                        "disk_offload",
                                        logger=log,
                                        module_name=resolved_label,
                                    ):
                                        offload_to_disk(
                                            model=looper.gptq_model.model,
                                            module=target_module,
                                            disk_path=offload_path,
                                        )
                                    if region_timer is not None and offload_start is not None:
                                        region_timer.record(
                                            "submodule_finalize_offload",
                                            time.perf_counter() - offload_start,
                                            source=resolved_label,
                                        )
                                else:
                                    log.warning(
                                        "Skipping disk offload for %s: no offload path configured",
                                        module_label,
                                    )
                    finally:
                        if region_timer is not None and start is not None:
                            region_timer.record(
                                "submodule_finalize",
                                time.perf_counter() - start,
                                source=resolved_label,
                            )
                    process_name = process.name() if process is not None else "<processor>"
                    return finalize_progress_cls(module_label, process_name, layer_idx)

                    # pb.subtitle(
                    #     f"{process.name()}: layer:{layer_idx} Finalized {idx}/{total} {module_label}"
                    # ).draw()

                for index, (process, module, module_label, target_dev, layer_idx) in enumerate(finalize_tasks, start=1):
                    # Schedule finalize work on the device thread pool so CPU
                    # bound tasks do not stall the main orchestration loop.
                    future = DEVICE_THREAD_POOL.submit(
                        target_dev,
                        _finalize_on_worker,
                        process,
                        module,
                        index,
                        finalize_count,
                        module_label,
                        layer_idx,
                    )
                    finalize_futures.append((future, index, module_label, process, layer_idx))

                finalize_futures_snapshot = list(finalize_futures)

                looper._emit_layer_complete(
                    layer_idx=layer_index,
                    submodule_finalized=False,
                    raise_in_place=True,
                )

                if finalize_futures_snapshot:
                    known_layers = sorted(
                        {
                            layer_idx
                            for _, _, _, _, layer_idx in finalize_futures_snapshot
                            if layer_idx is not None
                        }
                    )
                    includes_unknown = any(
                        layer_idx is None
                        for _, _, _, _, layer_idx in finalize_futures_snapshot
                    )

                    layer_heading = "Layer ?"
                    if known_layers:
                        sample_layers = ", ".join(str(idx) for idx in known_layers[:3])
                        if len(known_layers) > 3:
                            sample_layers += ", …"
                        suffix = ", ?" if includes_unknown else ""
                        prefix = "Layer" if len(known_layers) == 1 else "Layers"
                        layer_heading = f"{prefix} {sample_layers}{suffix}"
                    elif includes_unknown:
                        layer_heading = "Layer ?"

                    finalize_pb.title(
                        f"{layer_heading} Submodule finalize 0/{finalize_count}"
                    ).subtitle("Waiting for completions...").draw()

                def _drain_finalize_futures(
                    futures,
                    finalize_pb_local,
                    finalize_count_local,
                    layer_idx_for_callback,
                ):
                    completed_local = 0
                    try:
                        for future in as_completed(futures):
                            # Drain futures as they complete to surface errors
                            # quickly and keep the progress bar in sync.
                            try:
                                result = future.result()
                            except BaseException as exc:
                                log.exception("Submodule finalize task raised an exception")
                                looper._request_loop_stop(exc)
                                return

                            if isinstance(result, finalize_progress_cls):
                                module_label = result.module_label
                                process_name = result.process_name
                                layer_idx = result.layer_idx
                            elif isinstance(result, tuple) and len(result) == 3:
                                module_label, process_name, layer_idx = result
                            else:
                                module_label = None
                                process_name = "<processor>"
                                layer_idx = None

                            layer_label = f"Layer {layer_idx}" if layer_idx is not None else "Layer ?"
                            display_module = module_label or "<unnamed>"
                            subtitle = f"{process_name}: {display_module}"

                            completed_local += 1
                            finalize_pb_local.next()
                            finalize_pb_local.title(
                                f"{layer_label} Finalize {completed_local}/{finalize_count_local}"
                            ).subtitle(subtitle).draw()
                    finally:
                        finalize_pb_local.close()
                        looper._emit_layer_complete(
                            layer_idx=layer_idx_for_callback,
                            submodule_finalized=True,
                            raise_in_place=False,
                        )

                if finalize_futures_snapshot:
                    if looper.gptq_model.quantize_config.wait_for_submodule_finalizers:
                        # Synchronous: wait for all finalization to complete before proceeding to next layer
                        # This ensures all packing and writing tasks are done
                        _drain_finalize_futures(
                            [future for future, *_ in finalize_futures_snapshot],
                            finalize_pb,
                            finalize_count,
                            layer_index,
                        )
                        if looper.gptq_model.quantize_config.gc_mode == GcMode.ON_STAGE_END:
                            torch_empty_cache()
                    else:
                        # Asynchronous (current/default behavior): drain in background thread
                        # This allows next layer to start while current layer finalizes
                        finalizer_thread = threading.Thread(
                            target=_drain_finalize_futures,
                            args=(
                                [future for future, *_ in finalize_futures_snapshot],
                                finalize_pb,
                                finalize_count,
                                layer_index,
                            ),
                            name="SubmoduleFinalizeWatcher",
                            daemon=True,
                        )
                        looper.register_dangling_thread(finalizer_thread)
                        finalizer_thread.start()
                else:
                    looper._emit_layer_complete(
                        layer_idx=layer_index,
                        submodule_finalized=True,
                        raise_in_place=True,
                    )

        # Check for pause after completing each layer
        layer_info = f"layer {layer_index}" if not is_lm_head_module else "lm_head"
        looper.pause_controller.check_pause_point(f"after {layer_info}")
            
        # Unregister progress bar when moving to next layer
        looper.pause_controller.unregister_progress_bar(pb)



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from . import LlamaQModel


class XverseQModel(LlamaQModel):
    require_pkgs = ["transformers<=4.38.2", "tokenizers<=0.15.2"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import importlib

import accelerate
import torch


ipex_available = importlib.util.find_spec("intel_extension_for_pytorch") is not None
try:
    triton_available = True
except ImportError:
    triton_available = False



def get_module_by_name_suffix(model, module_name: str):
    for name, module in model.named_modules():
        if name.endswith(module_name):
            return module


def simple_dispatch_model(model, device_map):
    from accelerate.hooks import AlignDevicesHook, add_hook_to_module

    if "" in device_map:
        d = device_map[""]
        model = model.to(torch.device(d))
        model.hf_device_map = device_map
        return model

    tied_params = accelerate.utils.modeling.find_tied_parameters(model)
    if set(device_map.values()) == {"cpu"} or set(device_map.values()) == {
        "cpu",
        "disk",
    }:
        main_device = "cpu"
    else:
        main_device = [d for d in device_map.values() if d not in ["cpu", "disk"]][0]

    cpu_offload_group = [(n, d) for n, d in device_map.items() if d == "cpu"]
    prev_hook = None
    for idx, (n, d) in enumerate(cpu_offload_group):
        m = get_module_by_name_suffix(model, n)
        _, prev_hook = accelerate.cpu_offload_with_hook(
            m, execution_device=main_device, prev_module_hook=prev_hook
        )
    # set first cpu offload module's prev_module_hook to the last cpu offload module's hook
    if len(cpu_offload_group) > 1:
        get_module_by_name_suffix(
            model, cpu_offload_group[0][0]
        )._hf_hook.prev_module_hook = prev_hook

    for n, d in device_map.items():
        m = get_module_by_name_suffix(model, n)
        if d != "cpu":
            d = torch.device(d)
            hook = AlignDevicesHook(d, io_same_device=True, place_submodules=True)
            add_hook_to_module(m, hook)
    accelerate.utils.modeling.retie_parameters(model, tied_params)
    model.hf_device_map = device_map

    return model


def set_module_name(model, name, value):
    if "." in name:
        parent_name = name.rsplit(".", 1)[0]
        child_name = name[len(parent_name) + 1 :]
        parent = model.get_submodule(parent_name)
    else:
        parent_name = ""
        parent = model
        child_name = name

    setattr(parent, child_name, value)


def compute_memory_used_pct(device):
    memory_used = torch.cuda.max_memory_allocated(device) / (1024**3)
    memory_pct = (
        memory_used
        / (torch.cuda.get_device_properties(device).total_memory / (1024**3))
        * 100
    )
    return memory_pct


def get_best_device():
    if torch.backends.mps.is_available():
        return "mps"
    elif torch.cuda.is_available():
        return "cuda:0"
    elif torch.xpu.is_available():
        return "xpu:0"
    else:
        return "cpu"


def get_lowest_memory_device_index():
    device = None
    curr_device_memory_pct = 0
    for device_index in range(torch.cuda.device_count()):
        device_memory_pct = compute_memory_used_pct(device_index)
        if device is None or device_memory_pct < curr_device_memory_pct:
            device = device_index
            curr_device_memory_pct = device_memory_pct

    return device



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from __future__ import annotations

import math
from typing import Optional, Tuple

import torch

from .config import (
    FailSafe,
    QuantizeConfig,
    SmoothLog,
    SmoothMAD,
    SmoothOutlier,
    SmoothPercentile,
    SmoothPercentileAsymmetric,
    SmoothRowCol,
    SmoothSoftNorm,
)


# For Gaussian-like rows, raw MAD ~= 0.67449 * sigma. Normalize MAD so the
# configured `k` behaves like a sigma-width window instead of clipping more
# aggressively than intended.
MAD_TO_STD_SCALE = 1.4826


def _quantile(block: torch.Tensor, percentile: float) -> torch.Tensor:
    if percentile <= 0.0:
        return block.min(dim=1, keepdim=True).values
    if percentile >= 100.0:
        return block.max(dim=1, keepdim=True).values

    q = max(0.0, min(percentile / 100.0, 1.0))
    position = q * (block.shape[1] - 1)
    lower_index = int(math.floor(position))
    upper_index = int(math.ceil(position))
    lower = block.kthvalue(lower_index + 1, dim=1, keepdim=True).values

    if upper_index == lower_index:
        return lower

    upper = block.kthvalue(upper_index + 1, dim=1, keepdim=True).values
    return lower + (upper - lower) * (position - lower_index)


def _clamp_block(block: torch.Tensor, lo: torch.Tensor, hi: torch.Tensor) -> torch.Tensor:
    return torch.max(torch.min(block, hi), lo)


def smooth_block(
    block: torch.Tensor,
    failsafe: FailSafe,
    *,
    eps: float = 1e-8,
    group_size: Optional[int] = None,
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    method = getattr(failsafe, "smooth", None)
    if method is None:
        return block, None
    if group_size is not None and group_size < 0:
        group_size = block.shape[1]
    if group_size is not None and group_size < getattr(method, "group_size_threshold", 0):
        return block, None

    if isinstance(method, SmoothRowCol):
        axis = (method.axis or "row").lower()
        if axis == "col":
            col_rms = block.pow(2).mean(dim=0, keepdim=True).sqrt().clamp(min=eps)
            scale_factor = col_rms.mean().view(1, 1)
        else:
            scale_factor = block.pow(2).mean(dim=1, keepdim=True).sqrt().clamp(min=eps)
        return block / scale_factor, scale_factor

    block_f = block.float()

    if isinstance(method, SmoothSoftNorm):
        mean = block_f.mean(dim=1, keepdim=True)
        rms = (block_f - mean).pow(2).mean(dim=1, keepdim=True).sqrt().clamp(min=eps)
        k = float(method.k)
        block_norm = (block_f - mean) / rms
        block_norm = torch.clamp(block_norm, -k, k)
        return (block_norm * rms + mean).to(block.dtype), None

    if isinstance(method, SmoothPercentile):
        abs_block = block_f.abs()
        threshold = _quantile(abs_block, float(method.percentile))
        return torch.clamp(block_f, -threshold, threshold).to(block.dtype), None

    if isinstance(method, SmoothPercentileAsymmetric):
        low = float(method.low)
        high = float(method.high)
        lo = _quantile(block_f, low)
        hi = _quantile(block_f, high)
        return _clamp_block(block_f, lo, hi).to(block.dtype), None

    if isinstance(method, SmoothMAD):
        median = block_f.median(dim=1, keepdim=True).values
        mad = (block_f - median).abs().median(dim=1, keepdim=True).values * MAD_TO_STD_SCALE
        k = float(method.k)
        lo = median - k * mad
        hi = median + k * mad
        return _clamp_block(block_f, lo, hi).to(block.dtype), None

    if isinstance(method, SmoothOutlier):
        pct = float(method.pct)
        if pct <= 0.0:
            return block, None
        abs_block = block_f.abs()
        k = max(1, int(round(abs_block.shape[1] * (1.0 - pct / 100.0))))
        k = min(k, abs_block.shape[1])
        threshold = torch.kthvalue(abs_block, k, dim=1, keepdim=True).values
        return torch.clamp(block_f, -threshold, threshold).to(block.dtype), None

    if isinstance(method, SmoothLog):
        mu = max(float(method.mu), eps)
        abs_block = block_f.abs()
        log_mu = math.log1p(mu)
        log_vals = torch.log1p(abs_block * mu) / log_mu
        threshold = _quantile(log_vals, float(method.percentile))
        lin_threshold = (torch.exp(threshold * log_mu) - 1.0) / mu
        return torch.clamp(block_f, -lin_threshold, lin_threshold).to(block.dtype), None

    return block, None


def _eval_mse(block_f, min_val, max_val, base_zero, qcfg, maxq, shrink, eps):
    """Compute MSE for shrinkage factors [rows, n, 1]."""
    scale = torch.clamp((max_val.unsqueeze(1) * shrink - min_val.unsqueeze(1) * shrink) / maxq, min=eps)
    zero = base_zero.unsqueeze(1).expand_as(scale) if qcfg.sym else torch.round(-min_val.unsqueeze(1) * shrink / scale)
    q = torch.clamp(torch.round(block_f.unsqueeze(1) / scale + zero), 0, maxq)

    return ((q - zero) * scale - block_f.unsqueeze(1)).pow(2).mean(dim=2)


def mse_optimal_quant(
    block: torch.Tensor,
    qcfg: QuantizeConfig,
    maxq: int,
    *,
    steps: int,
    maxshrink: float,
    eps: float = 1e-8,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Ternary search: O(log steps)"""
    block_f = block.float()
    rows = block_f.shape[0]

    if qcfg.sym:
        max_abs = block_f.abs().max(dim=1, keepdim=True).values
        min_val = -max_abs
        max_val = max_abs
        base_zero = torch.full_like(max_abs, (maxq + 1) / 2.0)
    else:
        min_val = block_f.min(dim=1, keepdim=True).values
        max_val = block_f.max(dim=1, keepdim=True).values
        base_zero = None

    steps = max(int(math.log(max(steps, 2)) / math.log(1.5)) + 1, 3)
    shrink = max(min(maxshrink, 1.0), 1e-3)

    # Left, Right pointer
    l, r = torch.full((rows, 1), shrink, device=block_f.device, dtype=block_f.dtype), torch.ones((rows, 1), device=block_f.device, dtype=block_f.dtype)
    best_err, best_p = torch.full((rows,), float('inf'), device=block_f.device), r.clone()

    for _ in range(steps):
        mid1, mid2 = l + (r - l) / 3.0, r - (r - l) / 3.0
        err = _eval_mse(block_f, min_val, max_val, base_zero, qcfg, maxq, torch.stack([mid1, mid2], dim=1).view(rows, 2, 1), eps)

        for i, p in enumerate([mid1, mid2]):
            better = err[:, i] < best_err
            best_err, best_p = torch.where(better, err[:, i], best_err), torch.where(better.unsqueeze(1), p, best_p)

        move_r = err[:, 0] < err[:, 1]
        r, l = torch.where(move_r.unsqueeze(1), mid2, r), torch.where(move_r.unsqueeze(1), l, mid1)

    # Refine
    delta = (r - l) * 0.1
    refinement = torch.stack([torch.clamp(best_p - delta, shrink, 1.0), best_p, torch.clamp(best_p + delta, shrink, 1.0)], dim=1).view(rows, 3, 1)
    best_p = torch.gather(refinement.squeeze(2), 1, _eval_mse(block_f, min_val, max_val, base_zero, qcfg, maxq, refinement, eps).argmin(dim=1).unsqueeze(1))

    # Final quantization
    scale_best = torch.clamp((max_val - min_val) * best_p / maxq, min=eps)
    zero_best = base_zero if qcfg.sym else torch.round(-min_val * best_p / scale_best)
    q = torch.clamp(torch.round(block_f / scale_best + zero_best), 0, maxq)
    dequant_best = (q - zero_best) * scale_best

    return dequant_best, scale_best, zero_best



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from transformers import AutoModelForImageTextToText

from ..base import BaseQModel
from ..moe_lifecycle import GateUpDownMoELifecycleHooks


class Llama4QModel(BaseQModel):
    # some bug in the attention_mask of transformers.modeling_llama4,
    # so batch quantization for Llama4 is temporarily not supported.
    support_batch_quantize = False
    loader = AutoModelForImageTextToText

    pre_lm_head_norm_module = "language_model.model.norm"

    dynamic_expert_index = "num_local_experts"

    # MoE lifecycle hooks for gate_proj/up_proj/down_proj pattern
    moe_lifecycle_hooks = GateUpDownMoELifecycleHooks()

    module_tree = [
        "language_model",
        "model",
        "layers",
        "#",
        {
            "input_layernorm": ("input_layernorm:!",),
            "self_attn": ("q_proj:0", "k_proj:0", "v_proj:0", "o_proj:1"),
            "post_attention_layernorm": ("post_attention_layernorm:!",),
            "feed_forward:moe": {
                "experts": {
                    "#": ("gate_proj:0", "up_proj:0", "down_proj:1"),
                },
                "shared_expert": ("gate_proj:0", "up_proj:0", "down_proj:1"),
            },
        }
    ]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import threading
import time
import uuid

import torch


try:
    import uvicorn
    from fastapi import FastAPI, HTTPException
    from pydantic import BaseModel
except ModuleNotFoundError as exception:
    raise type(exception)(
        "GPTQModel OpenAi serve required dependencies are not installed.",
        "Please install via `pip install gptqmodel[openai] --no-build-isolation`.",
    )

class OpenAiServer:
    def __init__(self, model):
        self.uvicorn_server = None
        self.app = FastAPI()
        self.model = model
        self.tokenizer = model.tokenizer
        self.model_id_or_path = model.config.name_or_path
        self.setup_routes()

    def setup_routes(self):
        class OpenAiRequest(BaseModel):
            model: str
            messages: list = []
            max_tokens: int = 256
            temperature: float = 0.0
            top_p: float = 1.0
            n: int = 1
            stop: list = None

        class OpenAiResponseChoice(BaseModel):
            text: str
            index: int = 0

        class OpenAiResponse(BaseModel):
            id: str = ""
            object: str = "text_completion"
            created: int
            model: str
            choices: list[OpenAiResponseChoice]

        @self.app.post("/v1/chat/completions", response_model=OpenAiResponse)
        async def create_completion(request: OpenAiRequest):
            try:
                model_inputs = self.tokenizer.apply_chat_template(
                    request.messages,
                    add_generation_prompt=True,
                    return_tensors='pt',
                )

                if isinstance(model_inputs, torch.Tensor):
                    model_inputs = model_inputs.to(self.model.device)
                    generate_inputs = {"inputs": model_inputs}
                    prompt_length = model_inputs.size(-1)
                else:
                    model_inputs = model_inputs.to(self.model.device)
                    input_ids = model_inputs["input_ids"]
                    generate_inputs = dict(model_inputs)
                    prompt_length = input_ids.size(-1)

                do_sample = True if request.temperature != 0.0 else False
                with torch.inference_mode():
                    outputs = self.model.generate(
                        max_length=prompt_length + request.max_tokens,
                        temperature=request.temperature,
                        top_p=request.top_p,
                        num_return_sequences=request.n,
                        eos_token_id=self.tokenizer.eos_token_id,
                        stop_strings=request.stop,
                        do_sample=do_sample,
                        **generate_inputs,
                    )

                generated_texts = self.tokenizer.batch_decode(
                    outputs[:, prompt_length:],
                    skip_special_tokens=True,
                )

                choices = [
                    OpenAiResponseChoice(
                        text=gen_text,
                        index=i,
                    )
                    for i, gen_text in enumerate(generated_texts)
                ]

                response = OpenAiResponse(
                    id=f"{uuid.uuid4()}",
                    created=int(time.time()),
                    model=self.model_id_or_path,
                    choices=choices
                )
                return response
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        @self.app.get("/")
        def read_root():
            return {"message": "GPTQModel OpenAI Compatible Server is running."}

        @self.app.get("/shutdown")
        def shutdown():
            self.shutdown()
            return {"message": "Server is shutting down..."}

    def start(self, host: str = "0.0.0.0", port: int = 80, async_mode: bool = True):
        config = uvicorn.Config(self.app, host=host, port=port, log_level="info")
        self.uvicorn_server = uvicorn.Server(config)

        def run_server():
            self.uvicorn_server.run()

        if async_mode:
            thread = threading.Thread(target=run_server, daemon=False)
            thread.start()
            print(f"GPTQModel OpenAi Server has started asynchronously at http://{host}:{port}.")
        else:
            run_server()
            print(f"GPTQModel OpenAi Server has started synchronously at http://{host}:{port}.")

    def shutdown(self):
        if self.uvicorn_server is not None:
            self.uvicorn_server.should_exit = True
            print("GPTQModel OpenAi Server is shutting down...")

    def wait_until_ready(self, timeout: int = 30, check_interval: float = 0.1):
        start_time = time.time()
        while not self.uvicorn_server.started:
            if time.time() - start_time > timeout:
                raise TimeoutError("GPTQModel OpenAi server failed to start within the specified time.")
            time.sleep(check_interval)



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from __future__ import annotations

import contextlib
import threading

import torch


_GLOBAL_WARMUP_LOCK = threading.Lock()


def _make_spd(size: int, device: torch.device, dtype: torch.dtype) -> torch.Tensor:
    """Generate a small symmetric positive definite matrix."""
    base = torch.randn((size, size), device=device, dtype=dtype)
    identity = torch.eye(size, device=device, dtype=dtype)
    return base @ base.transpose(-1, -2) + identity * 1e-3


def _run_cholesky_and_eigh(device: torch.device, dtype: torch.dtype) -> None:
    spd = _make_spd(4, device, dtype)
    torch.linalg.cholesky(spd)

    # mps has no aten.eigh implementation
    if device.type == "mps":
        return

    torch.linalg.eigh(spd)


def _run_svd(device: torch.device, dtype: torch.dtype) -> None:
    mat = torch.randn((4, 3), device=device, dtype=dtype)
    torch.linalg.svd(mat, full_matrices=False)


def _run_qr(device: torch.device, dtype: torch.dtype) -> None:
    # mps has no aten.qr implementation
    if device.type == "mps":
        return

    square = torch.randn((4, 4), device=device, dtype=dtype)
    torch.linalg.qr(square)


def run_torch_linalg_warmup(device: torch.device) -> None:
    """
    Execute the torch.linalg operators used across the project once on the worker thread.

    Serialized under a global lock to avoid races inside PyTorch's lazy wrappers. The warmup
    still runs once per physical device so backend-specific handles are initialized where needed.
    """
    with _GLOBAL_WARMUP_LOCK:
        if device.type == "mps":
            dtypes = (torch.float32,)  # MPS backend does not implement float64.
        else:
            dtypes = (torch.float32, torch.float64)

        for dtype in dtypes:
            _run_cholesky_and_eigh(device, dtype)
            _run_svd(device, dtype)
            _run_qr(device, dtype)

        if device.type == "cuda" and hasattr(torch.backends, "cuda"):
            preferred = getattr(torch.backends.cuda, "preferred_linalg_library", None)
            if callable(preferred):
                current = preferred()
                # Core warmup already ran using the currently preferred backend above.
                # Some installations fall back to MAGMA when the primary solver is unavailable,
                # so we pre-initialize MAGMA as well when it differs from the preferred backend.
                if current and current != "magma":
                    with contextlib.suppress(Exception):
                        torch.backends.cuda.preferred_linalg_library(backend="magma")
                        _run_cholesky_and_eigh(device, torch.float32)
                if current:
                    with contextlib.suppress(Exception):
                        torch.backends.cuda.preferred_linalg_library(backend=current)


__all__ = ["run_torch_linalg_warmup"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import torch

from ..base import BaseQModel


class TeleChat2QModel(BaseQModel):
    # telechat2 requires custom model code
    require_trust_remote_code = True
    # telechat2 requires float16
    require_dtype = torch.float16

    pre_lm_head_norm_module = "transformer.ln_f"

    module_tree = [
        "transformer",
        "h",
        "#",
        {
            "input_layernorm": ("input_layernorm:!",),
            "self_attention": {"dense": ("dense",)},
            "post_attention_layernorm": ("post_attention_layernorm:!",),
            "mlp": ("gate_proj:0", "up_proj:0", "down_proj:1"),
        }
    ]

__all__ = ["TeleChat2QModel"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from ..base import BaseQModel


class BrumbyQModel(BaseQModel):
    require_trust_remote_code = True
    require_pkgs = ["retention>=1.0.7"]

    pre_lm_head_norm_module = "model.norm"

    module_tree = [
        "model",
        "layers",
        "#",
        {
            "input_layernorm": ("input_layernorm:0:!",),
            "self_attn": (
                "q_proj:0",
                "k_proj:0",
                "v_proj:0",
                "g_proj:0:!",
                "o_proj:1",
                "q_norm:0:!",
                "k_norm:0:!",
            ),
            "post_attention_layernorm": ("post_attention_layernorm:1:!",),
            "mlp": (
                "gate_proj:0",
                "up_proj:0",
                "down_proj:1",
            ),
        },
    ]

    def after_model_load(self, model, load_quantized_model=False):
        if hasattr(model, "config") and hasattr(model.config, "use_cache"):
            model.config.use_cache = False
        generation_config = getattr(model, "generation_config", None)
        if generation_config is not None and hasattr(generation_config, "use_cache"):
            generation_config.use_cache = False
        return model



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from collections import defaultdict
from typing import Iterable, List, Tuple

from .torch import TorchQuantLinear


def configure_lookahead_chain(modules: Iterable[TorchQuantLinear]):
    """Wire a sequence of TorchQuantLinear modules for one-step lookahead.

    Each module in *modules* (except the last) will prefetch the next module's
    dequantized weights the moment it finishes its own forward call. The last
    module's ``lookahead_next`` pointer is cleared.
    """

    last = None
    for module in modules:
        if last is not None:
            last.enable_lookahead(True).set_lookahead_next(module)
        module.enable_lookahead(True)
        last = module
    if last is not None:
        last.set_lookahead_next(None)


def _clear_existing_links(modules: Iterable[TorchQuantLinear]):
    for module in modules:
        module.set_lookahead_next(None)


def configure_default_lookahead(model) -> None:
    """Eagerly decode the MLP projection trio when attention ``q_proj`` runs.

    For each transformer block this disables lookahead between
    ``self_attn.{q,k,v,o}_proj`` and instead wires ``q_proj`` to prefetch the
    block's ``mlp.{gate,up,down}_proj`` modules concurrently. Missing modules
    are skipped.
    """

    ordered_modules: List[Tuple[str, TorchQuantLinear]] = []
    for name, module in model.named_modules():
        if isinstance(module, TorchQuantLinear):
            ordered_modules.append((name, module))

    if not ordered_modules:
        return

    _clear_existing_links(module for _, module in ordered_modules)

    attn_order = ("q_proj", "k_proj", "v_proj", "o_proj")
    mlp_order = ("gate_proj", "up_proj", "down_proj")

    attn_blocks = defaultdict(dict)
    mlp_blocks = defaultdict(dict)

    for name, module in ordered_modules:
        if ".self_attn." in name:
            prefix, leaf = name.split(".self_attn.", maxsplit=1)
            leaf = leaf.split(".")[0]
            attn_blocks[prefix][leaf] = module
            continue
        if ".mlp." in name:
            prefix, leaf = name.split(".mlp.", maxsplit=1)
            leaf = leaf.split(".")[0]
            mlp_blocks[prefix][leaf] = module

    for block in set(list(attn_blocks.keys()) + list(mlp_blocks.keys())):
        attn = attn_blocks.get(block, {})
        mlp = mlp_blocks.get(block, {})

        q_module = attn.get("q_proj")
        attn_modules = [attn.get(key) for key in attn_order if attn.get(key) is not None]
        mlp_targets = [mlp.get(key) for key in mlp_order if mlp.get(key) is not None]

        # reset lookahead state on all participating modules within this block
        for module in attn_modules:
            module.set_lookahead_next(None)
            module.enable_lookahead(False)
        for module in mlp_targets:
            module.set_lookahead_next(None)
            module.enable_lookahead(False)

        if q_module is not None and mlp_targets:
            q_module.enable_lookahead(True).set_lookahead_next(tuple(mlp_targets))
            for target in mlp_targets:
                target.enable_lookahead(True)



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import pytest
import torch
from gptqmodel_exllama_eora import gptq_gemm, gptq_gemm_eora

m = 1
k = 4096
n = 6144
r = 128

bit_default = 4
use_exllama = True

BLOCK_KN_SIZE=32
r_size = BLOCK_KN_SIZE * r / k


max_k1 = 16384
k_step1 = 128
input1 = [(k, r, bit_default) for k in range(k_step1, max_k1, k_step1) for r in range(k_step1, k, k_step1)]

max_k2 = 4096
k_step2 = 32
input2 = [(k, r, bits) for k in range(k_step2, max_k2, k_step2) for r in range(k_step2, k, k_step2) for bits in [2, 3, 4]]

#same as input 2 but r is not divisible by 32 (35, 67, etc)
input3 = [(k, r, bit_default) for k in range(k_step2, max_k2, k_step2) for r in range(k_step2 + 3, k, k_step2)]

input = input1 + input2 + input3

@pytest.mark.parametrize(
    "k, r, bit",
    input,
)
def test_eora_kernel_sizes(k, r, bit):
    x = torch.rand((m, k), device='cuda', dtype=torch.float16)
    eora_a = torch.randn((k, r), device='cuda', dtype=torch.float16) / 10.
    eora_b = torch.randn((r, n), device='cuda', dtype=torch.float16) / 10.

    ax = x @ eora_a

    gptq_groups = 32
    weight = torch.randint(-2000000, 2000000, (int(k / 2 / bit), n), device='cuda', dtype=torch.int32)
    zeros = torch.zeros((gptq_groups, int(n / 2 / bit)), device='cuda', dtype=torch.int32)
    scales = torch.rand((gptq_groups, n), device='cuda', dtype=torch.float16) / 1000.0
    idx = torch.empty((0,), device='cuda', dtype=torch.int32)

    gptq_pytorch_out = gptq_gemm(x, weight, zeros, scales, idx, use_exllama, bit) + (ax @ eora_b)
    gptq_eora_fused_out = gptq_gemm_eora(x, weight, zeros, scales, idx, use_exllama, bit, ax, eora_b)
    torch.testing.assert_close(gptq_pytorch_out, gptq_eora_fused_out, rtol=0.05, atol=1)  # 5 % relative tolerance, 1 absolute tolerance



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from __future__ import annotations

from typing import Optional, Tuple

import torch

from ...adapter.adapter import Adapter, Lora
from ...models._const import DEVICE, PLATFORM
from ...nn_modules.qlinear import AWQuantLinear
from ...quantization import FORMAT, METHOD
from ...utils.backend import BACKEND
from ...utils.logger import setup_logger
from ...utils.machete import (
    _validate_machete_device_support,
    gptqmodel_machete_kernels,
    machete_import_exception,
    machete_mm,
    machete_prepack_B,
    pack_quantized_values_into_int32,
)
from ...utils.marlin import replace_parameter, unpack_cols
from ...utils.marlin_scalar_type import scalar_types
from ...utils.rocm import IS_ROCM


log = setup_logger()


class AwqMacheteQuantLinear(AWQuantLinear):
    SUPPORTS_BACKENDS = [BACKEND.MACHETE]
    SUPPORTS_METHODS = [METHOD.AWQ]
    SUPPORTS_FORMATS = {FORMAT.GEMM: 100, FORMAT.MARLIN: 100}
    SUPPORTS_BITS = [4, 8]
    SUPPORTS_GROUP_SIZE = [-1, 32, 64, 128]
    SUPPORTS_DESC_ACT = [False]  # AWQ kernels do not reorder activations
    SUPPORTS_SYM = [True, False]
    SUPPORTS_SHARDS = True
    SUPPORTS_TRAINING = False
    SUPPORTS_AUTO_PADDING = False
    SUPPORTS_IN_FEATURES_DIVISIBLE_BY = [64]
    SUPPORTS_OUT_FEATURES_DIVISIBLE_BY = [128]

    SUPPORTS_DEVICES = [DEVICE.CUDA]
    SUPPORTS_PLATFORM = [PLATFORM.LINUX]
    SUPPORTS_PACK_DTYPES = [torch.int32]
    SUPPORTS_ADAPTERS = [Lora]

    SUPPORTS_DTYPES = [torch.float16, torch.bfloat16]

    REQUIRES_FORMAT_V2 = False

    QUANT_TYPE = "awq_machete"

    TYPE_MAP = {
        4: scalar_types.uint4,
        8: scalar_types.uint8,
    }

    def __init__(
            self,
            bits: int,
            group_size: int,
            desc_act: bool,
            sym: bool,
            in_features: int,
            out_features: int,
            bias: bool = False,
            pack_dtype: torch.dtype = torch.int32,
            adapter: Adapter = None,
            register_buffers: bool = False,
            **kwargs):
        if machete_import_exception is not None:
            raise ValueError(
                "Trying to use the machete backend, but could not import the "
                f"C++/CUDA dependencies with the following error: {machete_import_exception}"
            )

        if bits not in self.TYPE_MAP:
            raise ValueError(f"Unsupported num_bits = {bits}. Supported: {list(self.TYPE_MAP.keys())}")

        super().__init__(
            bits=bits,
            group_size=group_size,
            sym=sym,
            desc_act=False,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            pack_dtype=pack_dtype,
            backend=kwargs.pop("backend", BACKEND.MACHETE),
            adapter=adapter,
            register_buffers=register_buffers,
            **kwargs)

        self.weight_type = self.TYPE_MAP[self.bits]
        self.has_zero_points = True

    @classmethod
    def validate_once(cls) -> Tuple[bool, Optional[Exception]]:
        if gptqmodel_machete_kernels is None:
            return False, ImportError(machete_import_exception)
        else:
            return True, None

    @classmethod
    def validate_device(cls, device: DEVICE):
        super().validate_device(device)
        if device == DEVICE.CUDA:
            if IS_ROCM:
                raise NotImplementedError("Machete kernel is not supported on ROCm.")
            if not _validate_machete_device_support():
                raise NotImplementedError("Machete kernel requires compute capability >= 9.0.")

    def post_init(self):
        device = self.qweight.device

        # Reconstruct integer weights from packed AWQ representation
        qweight_int = unpack_cols(
            self.qweight,
            self.bits,
            self.in_features,
            self.out_features,
        ).to(device=device)

        packed = pack_quantized_values_into_int32(
            qweight_int,
            self.weight_type,
            packed_dim=0,
        )
        packed = packed.t().contiguous().t()
        prepacked = machete_prepack_B(
            packed,
            a_type=self.scales.dtype,
            b_type=self.weight_type,
            group_scales_type=self.scales.dtype,
        )
        replace_parameter(
            self,
            "qweight",
            torch.nn.Parameter(prepacked.contiguous(), requires_grad=False),
        )

        # Ensure scales are contiguous and resident on the correct device.
        replace_parameter(
            self,
            "scales",
            torch.nn.Parameter(self.scales.contiguous(), requires_grad=False),
        )

        # Convert zero-points: unpack columns, then pre-apply scales as expected by machete_mm
        effective_group_size = self.in_features if self.group_size == -1 else self.group_size
        num_groups = self.in_features // effective_group_size

        qzeros_unpacked = unpack_cols(
            self.qzeros,
            self.bits,
            num_groups,
            self.out_features,
        ).to(device=device)

        scales = self.scales
        qzeros_fp = (-1.0 * scales.to(dtype=scales.dtype) * qzeros_unpacked.to(scales.dtype)).contiguous()
        replace_parameter(
            self,
            "qzeros",
            torch.nn.Parameter(qzeros_fp, requires_grad=False),
        )

        if self.bias is not None:
            self.bias = self.bias.to(device=device)

        super().post_init()

    def forward(self, x: torch.Tensor):
        if x.shape[0] == 0:
            return torch.empty((0, self.out_features), dtype=x.dtype, device=x.device)

        input_2d = x.reshape(-1, x.shape[-1])
        group_scales = self.scales.to(dtype=input_2d.dtype)
        group_zeros = self.qzeros.to(dtype=input_2d.dtype)

        output = machete_mm(
            a=input_2d,
            b_q=self.qweight,
            b_type=self.weight_type,
            b_group_scales=group_scales,
            b_group_zeros=group_zeros,
            b_group_size=self.group_size,
        )

        if self.bias is not None:
            output.add_(self.bias)

        result = output.reshape(x.shape[:-1] + (self.out_features,))

        if self.adapter:
            result = self.adapter.apply(x=x, out=result)

        return result


__all__ = ["AwqMacheteQuantLinear"]
